from __future__ import (absolute_import, division, print_function)

from owslib.util import nspath_eval
from owslib.namespaces import Namespaces
from owslib.util import testXMLAttribute, testXMLValue, InfiniteDateTime, NegativeInfiniteDateTime

from dateutil import parser
from datetime import timedelta

from owslib.etree import etree

def get_namespaces():
    ns = Namespaces()
    return ns.get_namespaces(["swe20", "xlink"])
namespaces = get_namespaces()

def nspv(path):
    return nspath_eval(path, namespaces)

def make_pair(string, cast=None):
    if string is None:
        return None

    string = string.split(" ")
    if cast is not None:
        try:
            string = [cast(x) for x in string]
        except:
            print("Could not cast pair to correct type.  Setting to an empty tuple!")
            string = ""

    return tuple(string)

def get_uom(element):

    uom = testXMLAttribute(element, "code")
    if uom is None:
        uom = testXMLAttribute(element, nspv("xlink:href"))
    return uom

def get_boolean(value):
    if value is None:
        return None
    if value is True or value.lower() in ["yes","true"]:
        return True
    elif value is False or value.lower() in ["no","false"]:
        return False
    else:
        return None

def get_int(value):
    try:
        return int(value)
    except:
        return None

def get_float(value):
    try:
        return float(value)
    except:
        return None

AnyScalar    = [nspv(x) for x in ["swe20:Boolean", "swe20:Count", "swe20:Quantity", "swe20:Time", "swe20:Category", "swe20:Text"]]
AnyNumerical = [nspv(x) for x in ["swe20:Count", "swe20:Quantity", "swe20:Time"]]
AnyRange     = [nspv(x) for x in ["swe20:QuantityRange", "swe20:TimeRange", "swe20:CountRange", "swe20:CategoryRange"]]

class NamedObject(object):
    def __init__(self, element):
        # No call to super(), the type object will process that.
        self.name           = testXMLAttribute(element, "name")
        try:
            self.content    = eval(element[-1].tag.split("}")[-1])(element[-1])
        except IndexError:
            self.content    = None
        except BaseException:
            raise

    # Revert to the content if attribute does not exists
    def __getattr__(self, name):
        return getattr(self.content, name)

class AbstractSWE(object):
    def __init__(self, element):
        # Attributes
        self.id             = testXMLAttribute(element,"id")   # string, optional

        # Elements
        self.extention      = []                            # anyType, min=0, max=X

class AbstractSWEIdentifiable(AbstractSWE):
    def __init__(self, element):
        super(AbstractSWEIdentifiable, self).__init__(element)
        # Elements
        self.identifier     = testXMLValue(element.find(nspv("swe20:identifier")))    # anyURI, min=0
        self.label          = testXMLValue(element.find(nspv("swe20:label")))         # string, min=0
        self.description    = testXMLValue(element.find(nspv("swe20:description")))   # string, min=0

class AbstractDataComponent(AbstractSWEIdentifiable):
    def __init__(self, element):
        super(AbstractDataComponent, self).__init__(element)
        # Attributes
        self.definition     = testXMLAttribute(element,"definition")                        # anyURI, required
        self.updatable      = get_boolean(testXMLAttribute(element,"updatable"))            # boolean, optional
        self.optional       = get_boolean(testXMLAttribute(element,"optional")) or False    # boolean, default=False

class AbstractSimpleComponent(AbstractDataComponent):
    def __init__(self, element):
        super(AbstractSimpleComponent, self).__init__(element)
        # Attributes
        self.referenceFrame = testXMLAttribute(element,"referenceFrame")    # anyURI, optional
        self.axisID         = testXMLAttribute(element,"axisID")            # string, optional

        # Elements
        self.quality        = [_f for _f in [Quality(q) for q in [e.find('*') for e in element.findall(nspv("swe20:quality"))] if q is not None] if _f]
        try:
            self.nilValues  = NilValues(element.find(nspv("swe20:nilValues")))
        except:
            self.nilValues  = None

class Quality(object):
    def __new__(cls, element):
        t = element.tag.split("}")[-1]
        if t == "Quantity":
            return Quantity(element)
        elif t == "QuantityRange":
            return QuantityRange(element)
        elif t == "Category":
            return Category(element)
        elif t == "Text":
            return Text(element)
        else:
            return None

class NilValues(AbstractSWE):
    def __init__(self, element):
        super(NilValues, self).__init__(element)
        self.nilValue           = [_f for _f in [nilValue(x) for x in element.findall(nspv("swe20:nilValue"))] if _f] # string, min=0, max=X

class nilValue(object):
    def __init__(self, element):
        self.reason             = testXMLAttribute(element, "reason")
        self.value              = testXMLValue(element)

class AllowedTokens(AbstractSWE):
    def __init__(self, element):
        super(AllowedTokens, self).__init__(element)
        self.value              = [_f for _f in [testXMLValue(x) for x in element.findall(nspv("swe20:value"))] if _f] # string, min=0, max=X
        self.pattern            = testXMLValue(element.find(nspv("swe20:pattern")))                             # string (Unicode Technical Standard #18, Version 13), min=0

class AllowedValues(AbstractSWE):
    def __init__(self, element):
        super(AllowedValues, self).__init__(element)
        self.value              = [_f for _f in [get_float(x) for x in [testXMLValue(x) for x in element.findall(nspv("swe20:value"))]] if _f]
        self.interval           = [_f for _f in [make_pair(testXMLValue(x)) for x in element.findall(nspv("swe20:interval"))] if _f]
        self.significantFigures = get_int(testXMLValue(element.find(nspv("swe20:significantFigures"))))                                         # integer, min=0

class AllowedTimes(AbstractSWE):
    def __init__(self, element):
        super(AllowedTimes, self).__init__(element)
        self.value              = [_f for _f in [testXMLValue(x) for x in element.findall(nspv("swe20:value"))] if _f]
        self.interval           = [_f for _f in [make_pair(testXMLValue(x)) for x in element.findall(nspv("swe20:interval"))] if _f]
        self.significantFigures = get_int(testXMLValue(element.find(nspv("swe20:significantFigures"))))                         # integer, min=0

class Boolean(AbstractSimpleComponent):
    def __init__(self, element):
        super(Boolean, self).__init__(element)
        # Elements
        """
        6.2.1 Boolean
                A Boolean representation of a proptery can take only two values that should be "true/false" or "yes/no".
        """
        value               = get_boolean(testXMLValue(element.find(nspv("swe20:value"))))   # boolean, min=0, max=1

class Text(AbstractSimpleComponent):
    def __init__(self, element):
        super(Text, self).__init__(element)
        # Elements
        """
        Req 6. A textual representation shall at least consist of a character string.
        """
        self.value          = testXMLValue(element.find(nspv("swe20:value")))                               # string, min=0, max=1

        try:
            self.constraint     = AllowedTokens(element.find(nspv("swe20:constraint/swe20:AllowedTokens"))) # AllowedTokens, min=0, max=1
        except:
            self.constraint     = None


class Category(AbstractSimpleComponent):
    def __init__(self, element):
        super(Category, self).__init__(element)
        # Elements
        self.codeSpace      = testXMLAttribute(element.find(nspv("swe20:codeSpace")), nspv("xlink:href"))   # Reference, min=0, max=1
        self.value          = testXMLValue(element.find(nspv("swe20:value")))                               # string, min=0, max=1

        try:
            self.constraint     = AllowedTokens(element.find(nspv("swe20:constraint/swe20:AllowedTokens"))) # AllowedTokens, min=0, max=1
        except:
            self.constraint     = None


class CategoryRange(Category):
    def __init__(self, element):
        super(CategoryRange, self).__init__(element)
        # Elements
        value               = testXMLValue(element.find(nspv("swe20:value")))
        self.values         = make_pair(value) if value is not None else None

class Count(AbstractSimpleComponent):
    def __init__(self, element):
        super(Count, self).__init__(element)
        # Elements
        self.value          = get_int(testXMLValue(element.find(nspv("swe20:value"))))                  # integer, min=0, max=1

        try:
            self.constraint = AllowedValues(element.find(nspv("swe20:constraint/swe20:AllowedValues"))) # AllowedValues, min=0, max=1
        except:
            self.constraint = None


class CountRange(Count):
    def __init__(self, element):
        super(CountRange, self).__init__(element)
        # Elements
        value               = testXMLValue(element.find(nspv("swe20:value")))
        self.value          = make_pair(value,int) if value is not None else None

class Quantity(AbstractSimpleComponent):
    def __init__(self, element):
        super(Quantity, self).__init__(element)
        # Elements
        self.uom            = get_uom(element.find(nspv("swe20:uom")))
        self.value          = get_float(testXMLValue(element.find(nspv("swe20:value"))))                  # double, min=0, max=1

        try:
            self.constraint = AllowedValues(element.find(nspv("swe20:constraint/swe20:AllowedValues")))   # AllowedValues, min=0, max=1
        except:
            self.constraint = None

class QuantityRange(Quantity):
    def __init__(self, element):
        super(QuantityRange, self).__init__(element)
        # Elements
        value               = testXMLValue(element.find(nspv("swe20:value")))
        self.value          = make_pair(value,float) if value is not None else None

def get_time(value, referenceTime, uom):
    try:
        value = parser.parse(value)

    except (AttributeError, ValueError): # Most likely an integer/float using a referenceTime
        try:
            if uom.lower() == "s":
                value  = referenceTime + timedelta(seconds=float(value))
            elif uom.lower() == "min":
                value  = referenceTime + timedelta(minutes=float(value))
            elif uom.lower() == "h":
                value  = referenceTime + timedelta(hours=float(value))
            elif uom.lower() == "d":
                value  = referenceTime + timedelta(days=float(value))

        except (AttributeError, ValueError):
            pass

    except OverflowError: # Too many numbers (> 10) or INF/-INF
        if value.lower() == "inf":
            value  = InfiniteDateTime()
        elif value.lower() == "-inf":
            value  = NegativeInfiniteDateTime()

    return value

class Time(AbstractSimpleComponent):
    def __init__(self, element):
        super(Time, self).__init__(element)
        # Elements
        self.uom                = get_uom(element.find(nspv("swe20:uom")))

        try:
            self.constraint     = AllowedTimes(element.find(nspv("swe20:constraint/swe20:AllowedTimes"))) # AllowedTimes, min=0, max=1
        except:
            self.constraint     = None

        # Attributes
        self.localFrame         = testXMLAttribute(element,"localFrame")                                    # anyURI, optional
        try:
            self.referenceTime  = parser.parse(testXMLAttribute(element,"referenceTime"))                   # dateTime, optional
        except (AttributeError, ValueError):
            self.referenceTime  = None

        value                   = testXMLValue(element.find(nspv("swe20:value")))                           # TimePosition, min=0, max=1
        self.value              = get_time(value, self.referenceTime, self.uom)

class TimeRange(AbstractSimpleComponent):
    def __init__(self, element):
        super(TimeRange, self).__init__(element)
        # Elements
        self.uom                = get_uom(element.find(nspv("swe20:uom")))

        try:
            self.constraint     = AllowedTimes(element.find(nspv("swe20:constraint/swe20:AllowedTimes"))) # AllowedTimes, min=0, max=1
        except:
            self.constraint     = None

        # Attributes
        self.localFrame         = testXMLAttribute(element,"localFrame")                                # anyURI, optional
        try:
            self.referenceTime  = parser.parse(testXMLAttribute(element,"referenceTime"))               # dateTime, optional
        except (AttributeError, ValueError):
            self.referenceTime  = None

        values                  = make_pair(testXMLValue(element.find(nspv("swe20:value"))))            # TimePosition, min=0, max=1
        self.value              = [get_time(t, self.referenceTime, self.uom) for t in values]

class DataRecord(AbstractDataComponent):
    def __init__(self, element):
        super(DataRecord, self).__init__(element)
        # Elements
        self.field          = [Field(x) for x in element.findall(nspv("swe20:field"))]
    def get_by_name(self, name):
        return next((x for x in self.field if x.name == name), None)

class Field(NamedObject):
    def __init__(self, element):
        super(Field, self).__init__(element)

class Vector(AbstractDataComponent):
    def __init__(self, element):
        super(Vector, self).__init__(element)
        # Elements
        self.coordinate     = [Coordinate(x) for x in element.findall(nspv("swe20:coordinate"))]

        # Attributes
        self.referenceFrame = testXMLAttribute(element,"referenceFrame")        # anyURI, required
        self.localFrame     = testXMLAttribute(element,"localFrame")            # anyURI, optional
    def get_by_name(self, name):
        return next((x for x in self.coordinate if x.name == name), None)

class Coordinate(NamedObject):
    def __init__(self, element):
        super(Coordinate, self).__init__(element)
        #if element[-1].tag not in AnyNumerical:
        #    print "Coordinate does not appear to be an AnyNumerical member"

class DataChoice(AbstractDataComponent):
    def __init__(self, element):
        super(DataChoice, self).__init__(element)
        self.item           = [Item(x) for x in element.findall(nspv("swe20:item"))]
    def get_by_name(self, name):
        return next((x for x in self.item if x.name == name), None)

class Item(NamedObject):
    def __init__(self, element):
        super(Item, self).__init__(element)

class DataArray(AbstractDataComponent):
    def __init__(self, element):
        super(DataArray, self).__init__(element)
        self.elementCount   = element.find(nspv("swe20:elementCount/swe20:Count"))      # required
        self.elementType    = ElementType(element.find(nspv("swe20:elementType")))      # required
        self.values         = testXMLValue(element.find(nspv("swe20:values")))
        try:
            self.encoding   = AbstractEncoding(element.find(nspv("swe20:encoding")))
        except:
            self.encoding   = None

class Matrix(AbstractDataComponent):
    def __init__(self, element):
        super(Matrix, self).__init__(element)
        self.elementCount   = element.find(nspv("swe20:elementCount/swe20:Count"))      # required
        self.elementType    = ElementType(element.find(nspv("swe20:elementType")))      # required
        self.encoding       = AbstractEncoding(element.find(nspv("swe20:encoding")))
        self.values         = testXMLValue(element.find(nspv("swe20:values")))
        self.referenceFrame = testXMLAttribute(element, "referenceFrame")               # anyURI, required
        self.localFrame     = testXMLAttribute(element, "localFrame")                   # anyURI, optional

class DataStream(AbstractSWEIdentifiable):
    def __init__(self, element):
        super(DataStream, self).__init__(element)
        self.elementCount   = element.find(nspv("swe20:elementCount/swe20:Count"))      # optional
        self.elementType    = ElementType(element.find(nspv("swe20:elementType")))      # optional
        self.encoding       = AbstractEncoding(element.find(nspv("swe20:encoding")))
        self.values         = testXMLValue(element.find(nspv("swe20:values")))

class ElementType(NamedObject):
    def __init__(self, element):
        super(ElementType, self).__init__(element)

class AbstractEncoding(object):
    def __new__(cls, element):
        t = element[-1].tag.split("}")[-1]
        if t == "TextEncoding":
            return super(AbstractEncoding, cls).__new__(TextEncoding)
        elif t == "XMLEncoding":
            return super(AbstractEncoding, cls).__new__(XMLEncoding)
        elif t == "BinaryEncoding":
            return super(AbstractEncoding, cls).__new__(BinaryEncoding)

class TextEncoding(AbstractEncoding):
    def __init__(self, element):
        self.tokenSeparator         = testXMLAttribute(element[-1], "tokenSeparator")                           # string,  required
        self.blockSeparator         = testXMLAttribute(element[-1], "blockSeparator")                           # string,  required
        self.decimalSeparator       = testXMLAttribute(element[-1], "decimalSeparator") or "."                  # string,  optional, default="."
        self.collapseWhiteSpaces    = get_boolean(testXMLAttribute(element[-1], "collapseWhiteSpaces")) or True # boolean, optional, default=True

class XMLEncoding(AbstractEncoding):
    def __init__(self, element):
        raise NotImplementedError

class BinaryEncoding(AbstractEncoding):
    def __init__(self, element):
        raise NotImplementedError



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   12/18/2004
#
#------------------------------------------------------------------------------

""" Defines the base class for the wxPython-based Traits UI modal and non-modal
    dialogs.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

from traits.api \
    import HasStrictTraits, HasPrivateTraits, Instance, List, Event

from traitsui.api \
    import UI

from traitsui.menu \
    import Action

from editor \
    import Editor

#-------------------------------------------------------------------------
#  Constants:
#-------------------------------------------------------------------------

# List of all predefined system button names:
SystemButtons = ['Undo', 'Redo', 'Apply', 'Revert', 'OK', 'Cancel', 'Help']

# List of alternative context items that might handle an Action 'perform':
PerformHandlers = ('object', 'model')

#-------------------------------------------------------------------------
#  'RadioGroup' class:
#-------------------------------------------------------------------------


class RadioGroup(HasStrictTraits):
    """ A group of mutually-exclusive menu or toolbar actions.
    """
    # List of menu or tool bar items
    items = List

    #-------------------------------------------------------------------------
    #  Handles a menu item in the group being checked:
    #-------------------------------------------------------------------------

    def menu_checked(self, menu_item):
        """ Handles a menu item in the group being checked.
        """
        for item in self.items:
            if item is not menu_item:
                item.control.Check(False)
                item.item.action.checked = False

    #-------------------------------------------------------------------------
    #  Handles a tool bar item in the group being checked:
    #-------------------------------------------------------------------------

    def toolbar_checked(self, toolbar_item):
        """ Handles a tool bar item in the group being checked.
        """
        for item in self.items:
            if item is not toolbar_item:
                item.tool_bar.ToggleTool(item.control_id, False)
                item.item.action.checked = False

#-------------------------------------------------------------------------
#  'ButtonEditor' class:
#-------------------------------------------------------------------------


class ButtonEditor(Editor):
    """ Editor for buttons.
    """
    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Action associated with the button
    action = Instance(Action)

    #-------------------------------------------------------------------------
    #  Initializes the object:
    #-------------------------------------------------------------------------

    def __init__(self, **traits):
        self.set(**traits)

    #-------------------------------------------------------------------------
    #  Handles the associated button being clicked:
    #-------------------------------------------------------------------------

    def perform(self, event):
        """ Handles the associated button being clicked.
        """
        self.ui.do_undoable(self._perform, event)

    def _perform(self, event):
        method_name = self.action.action
        if method_name == '':
            method_name = '_%s_clicked' % (self.action.name.lower())
        method = getattr(self.ui.handler, method_name, None)
        if method is not None:
            method(self.ui.info)
        else:
            self.action.perform(event)

#-------------------------------------------------------------------------
#  'BaseDialog' class:
#-------------------------------------------------------------------------


class BaseDialog(object):
    """ Base class for Traits UI dialog boxes.
    """
    #-------------------------------------------------------------------------
    #  Sets the frame's icon:
    #-------------------------------------------------------------------------

    def set_icon(self, icon=None):
        """ Sets the frame's icon.
        """
        from pyface.image_resource import ImageResource

        if not isinstance(icon, ImageResource):
            icon = ImageResource('frame.ico')
        self.control.SetIcon(icon.create_icon())

    #-------------------------------------------------------------------------
    #  Adds a status bar to the dialog:
    #-------------------------------------------------------------------------

    def add_statusbar(self):
        """ Adds a status bar to the dialog.
        """
        ui = self.ui
        statusbar = ui.view.statusbar
        context = ui.context
        if statusbar is not None:
            widths = []
            listeners = []
            control = wx.StatusBar(self.control)
            control.SetFieldsCount(len(statusbar))
            for i, item in enumerate(statusbar):
                width = abs(item.width)
                if width <= 1.0:
                    widths.append(-max(1, int(1000 * width)))
                else:
                    widths.append(int(width))

                set_text = self._set_status_text(control, i)
                name = item.name
                set_text(ui.get_extended_value(name))
                col = name.find('.')
                object = 'object'
                if col >= 0:
                    object = name[: col]
                    name = name[col + 1:]
                object = context[object]
                object.on_trait_change(set_text, name, dispatch='ui')
                listeners.append((object, set_text, name))

            control.SetStatusWidths(widths)
            self.control.SetStatusBar(control)
            ui._statusbar = listeners

    def _set_status_text(self, control, i):
        def set_status_text(text):
            control.SetStatusText(text, i)

        return set_status_text

    #-------------------------------------------------------------------------
    #  Adds a menu bar to the dialog:
    #-------------------------------------------------------------------------

    def add_menubar(self):
        """ Adds a menu bar to the dialog.
        """
        menubar = self.ui.view.menubar
        if menubar is not None:
            self._last_group = self._last_parent = None
            self.control.SetMenuBar(
                menubar.create_menu_bar(self.control, self))
            self._last_group = self._last_parent = None

    #-------------------------------------------------------------------------
    #  Adds a tool bar to the dialog:
    #-------------------------------------------------------------------------

    def add_toolbar(self):
        """ Adds a toolbar to the dialog.
        """
        toolbar = self.ui.view.toolbar
        if toolbar is not None:
            self._last_group = self._last_parent = None
            self.control.SetToolBar(
                toolbar.create_tool_bar(self.control, self))
            self._last_group = self._last_parent = None

    #-------------------------------------------------------------------------
    #  Adds a menu item to the menu bar being constructed:
    #-------------------------------------------------------------------------

    def add_to_menu(self, menu_item):
        """ Adds a menu item to the menu bar being constructed.
        """
        item = menu_item.item
        action = item.action

        if action.id != '':
            self.ui.info.bind(action.id, menu_item)

        if action.style == 'radio':
            if ((self._last_group is None) or
                    (self._last_parent is not item.parent)):
                self._last_group = RadioGroup()
                self._last_parent = item.parent
            self._last_group.items.append(menu_item)
            menu_item.group = self._last_group

        if action.enabled_when != '':
            self.ui.add_enabled(action.enabled_when, menu_item)

        if action.checked_when != '':
            self.ui.add_checked(action.checked_when, menu_item)

    #-------------------------------------------------------------------------
    #  Adds a tool bar item to the tool bar being constructed:
    #-------------------------------------------------------------------------

    def add_to_toolbar(self, toolbar_item):
        """ Adds a toolbar item to the toolbar being constructed.
        """
        self.add_to_menu(toolbar_item)

    #-------------------------------------------------------------------------
    #  Returns whether the menu action should be defined in the user interface:
    #-------------------------------------------------------------------------

    def can_add_to_menu(self, action, action_event=None):
        """ Returns whether the action should be defined in the user interface.
        """
        if action.defined_when == '':
            return True

        return self.ui.eval_when(action.defined_when)

    #-------------------------------------------------------------------------
    #  Returns whether the toolbar action should be defined in the user
    #  interface:
    #-------------------------------------------------------------------------

    def can_add_to_toolbar(self, action):
        """ Returns whether the toolbar action should be defined in the user
            interface.
        """
        return self.can_add_to_menu(action)

    #-------------------------------------------------------------------------
    #  Performs the action described by a specified Action object:
    #-------------------------------------------------------------------------

    def perform(self, action):
        """ Performs the action described by a specified Action object.
        """
        self.ui.do_undoable(self._perform, action)

    def _perform(self, action):
        method = getattr(self.ui.handler, action.action, None)
        if method is not None:
            method(self.ui.info)
        else:
            context = self.ui.context
            for item in PerformHandlers:
                handler = context.get(item, None)
                if handler is not None:
                    method = getattr(handler, action.action, None)
                    if method is not None:
                        method()
                        break
            else:
                action.perform()

    #-------------------------------------------------------------------------
    #  Check to see if a specified 'system' button is in the buttons list, and
    # add it if it is not:
    #-------------------------------------------------------------------------

    def check_button(self, buttons, action):
        """ Adds *action* to the system buttons list for this dialog, if it is
        not already in the list.
        """
        name = action.name
        for button in buttons:
            if self.is_button(button, name):
                return
        buttons.append(action)

    #-------------------------------------------------------------------------
    #  Check to see if a specified Action button is a 'system' button:
    #-------------------------------------------------------------------------

    def is_button(self, action, name):
        """ Returns whether a specified action button is a system button.
        """
        if isinstance(action, basestring):
            return (action == name)
        return (action.name == name)

    #-------------------------------------------------------------------------
    #  Coerces a string to an Action if necessary:
    #-------------------------------------------------------------------------

    def coerce_button(self, action):
        """ Coerces a string to an Action if necessary.
        """
        if isinstance(action, basestring):
            return Action(name=action,
                          action='?'[(not action in SystemButtons):])
        return action

    #-------------------------------------------------------------------------
    #  Creates a user specified button:
    #-------------------------------------------------------------------------

    def add_button(self, action, sizer, method=None, enabled=True,
                   name=None, default=False):
        """ Creates a button.
        """
        ui = self.ui
        if ((action.defined_when != '') and
                (not ui.eval_when(action.defined_when))):
            return None

        if name is None:
            name = action.name
        id = action.id
        button = wx.Button(self.control, -1, name)
        button.Enable(enabled)
        if default:
            button.SetDefault()
        if (method is None) or (action.enabled_when != '') or (id != ''):
            editor = ButtonEditor(ui=ui,
                                  action=action,
                                  control=button)
            if id != '':
                ui.info.bind(id, editor)
            if action.visible_when != '':
                ui.add_visible(action.visible_when, editor)
            if action.enabled_when != '':
                ui.add_enabled(action.enabled_when, editor)
            if method is None:
                method = editor.perform
        wx.EVT_BUTTON(self.control, button.GetId(), method)
        sizer.Add(button, 0, wx.LEFT, 5)
        if action.tooltip != '':
            button.SetToolTipString(action.tooltip)
        return button



from __future__ import absolute_import

import os
import sys

from traits.testing.unittest_tools import unittest

from ..python_editor import PythonEditor
from ..toolkit import toolkit_object
from ..window import Window

GuiTestAssistant = toolkit_object('util.gui_test_assistant:GuiTestAssistant')
no_gui_test_assistant = (GuiTestAssistant.__name__ == 'Unimplemented')


PYTHON_SCRIPT = os.path.join(os.path.dirname(__file__), 'python_shell_script.py')


@unittest.skipIf(no_gui_test_assistant, 'No GuiTestAssistant')
class TestPythonEditor(unittest.TestCase, GuiTestAssistant):

    def setUp(self):
        GuiTestAssistant.setUp(self)
        self.window = Window()
        self.window._create()

    def tearDown(self):
        if self.widget.control is not None:
            with self.delete_widget(self.widget.control):
                self.widget.destroy()
        if self.window.control is not None:
            with self.delete_widget(self.window.control):
                self.window.destroy()
        GuiTestAssistant.tearDown(self)

    def test_lifecycle(self):
        # test that destroy works
        self.widget = PythonEditor(self.window.control)
        self.gui.process_events()
        self.assertFalse(self.widget.dirty)
        self.widget.destroy()
        self.gui.process_events()

    def test_show_line_numbers(self):
        # test that destroy works
        self.widget = PythonEditor(self.window.control, show_line_numbers=False)
        self.gui.process_events()
        self.widget.show_line_numbers = True
        self.gui.process_events()
        self.widget.show_line_numbers = False
        self.gui.process_events()
        self.widget.destroy()
        self.gui.process_events()

    def test_load(self):
        # test that destroy works
        self.widget = PythonEditor(self.window.control)
        self.gui.process_events()
        with self.assertTraitChanges(self.widget, 'changed', count=1):
            self.widget.path = PYTHON_SCRIPT
        self.assertFalse(self.widget.dirty)
        self.widget.destroy()
        self.gui.process_events()

    def test_select_line(self):
        # test that destroy works
        self.widget = PythonEditor(self.window.control, path=PYTHON_SCRIPT)
        self.gui.process_events()
        self.widget.select_line(3)
        self.widget.destroy()
        self.gui.process_events()



#!/usr/bin/env python
from fs.tests import FSTestCases, ThreadingTestCases

import unittest

import os
import sys
import shutil
import tempfile
import subprocess
import time
from os.path import abspath
import urllib

from six import PY3


try:
    from pyftpdlib.authorizers import DummyAuthorizer
    from pyftpdlib.handlers import FTPHandler
    from pyftpdlib.servers import FTPServer
except ImportError:
    if not PY3:
        raise ImportError("Requires pyftpdlib <https://github.com/giampaolo/pyftpdlib>")

from fs.path import *

from fs import ftpfs

ftp_port = 30000
class TestFTPFS(unittest.TestCase, FSTestCases, ThreadingTestCases):

    __test__ = not PY3

    def setUp(self):
        global ftp_port
        ftp_port += 1
        use_port = str(ftp_port)
        #ftp_port = 10000
        self.temp_dir = tempfile.mkdtemp(u"ftpfstests")

        file_path = __file__
        if ':' not in file_path:
            file_path = abspath(file_path)
        # Apparently Windows requires values from default environment, so copy the exisiting os.environ
        env = os.environ.copy()
        env['PYTHONPATH'] = os.getcwd() + os.pathsep + env.get('PYTHONPATH', '')
        self.ftp_server = subprocess.Popen([sys.executable,
                                            file_path,
                                            self.temp_dir,
                                            use_port],
                                           stdout=subprocess.PIPE,
                                           env=env)
        # Block until the server writes a line to stdout
        self.ftp_server.stdout.readline()

        # Poll until a connection can be made
        start_time = time.time()
        while time.time() - start_time < 5:
            try:
                ftpurl = urllib.urlopen('ftp://127.0.0.1:%s' % use_port)
            except IOError:
                time.sleep(0)
            else:
                ftpurl.read()
                ftpurl.close()
                break
        else:
            # Avoid a possible infinite loop
            raise Exception("Unable to connect to ftp server")

        self.fs = ftpfs.FTPFS('127.0.0.1', 'user', '12345', dircache=True, port=use_port, timeout=5.0)
        self.fs.cache_hint(True)


    def tearDown(self):
        #self.ftp_server.terminate()
        if sys.platform == 'win32':
            os.popen('TASKKILL /PID '+str(self.ftp_server.pid)+' /F')
        else:
            os.system('kill '+str(self.ftp_server.pid))
        shutil.rmtree(self.temp_dir)
        self.fs.close()

    def check(self, p):
        check_path = self.temp_dir.rstrip(os.sep) + os.sep + p
        return os.path.exists(check_path.encode('utf-8'))


if __name__ == "__main__":

    # Run an ftp server that exposes a given directory
    import sys
    authorizer = DummyAuthorizer()
    authorizer.add_user("user", "12345", sys.argv[1], perm="elradfmw")
    authorizer.add_anonymous(sys.argv[1])

    #def nolog(*args):
    #    pass
    #ftpserver.log = nolog
    #ftpserver.logline = nolog

    handler = FTPHandler
    handler.authorizer = authorizer
    address = ("127.0.0.1", int(sys.argv[2]))
    #print address

    ftpd = FTPServer(address, handler)

    sys.stdout.write('serving\n')
    sys.stdout.flush()
    ftpd.serve_forever()



# Enthought library imports.
from traits.api import HasTraits, Instance

# Local imports.
from task_pane import TaskPane


class TraitsTaskPane(TaskPane):
    """ A TaskPane that displays a Traits UI View.
    """

    #### TraitsTaskPane interface #############################################

    # The model object to view. If not specified, the pane is used instead.
    model = Instance(HasTraits)

    # The UI object associated with the Traits view, if it has been constructed.
    ui = Instance('traitsui.ui.UI')

    ###########################################################################
    # 'HasTraits' interface.
    ###########################################################################

    def trait_context(self):
        """ Use the model object for the Traits UI context, if appropriate.
        """
        if self.model:
            return { 'object': self.model, 'pane': self }
        return super(TraitsTaskPane, self).trait_context()

    ###########################################################################
    # 'ITaskPane' interface.
    ###########################################################################

    def create(self, parent):
        """ Create and set the toolkit-specific control that represents the
            pane.
        """
        self.ui = self.edit_traits(kind='subpanel', parent=parent)
        self.control = self.ui.control

    def destroy(self):
        """ Destroy the toolkit-specific control that represents the pane.
        """
        self.ui.dispose()
        self.control = self.ui = None



# setup.py requires that these be defined, and the OnceAndOnlyOnce
# principle is used here.  This is the only place where these values
# are defined in the source distribution, and everything else that
# needs this should grab it from here.
__author__ = "Rob McMullen"
__author_email__ = "feedback@playermissile.com"
__url__ = "http://playermissile.com/omnivore"
__download_url__ = "https://github.com/robmcmullen/omnivore/archive"
__bug_report_url__ = "https://github.com/robmcmullen/omnivore/issues"

# The real version number is maintained in a file that's under version control
# so I don't have to keep updating and checking in the file
try:
    import _omnivore_version
    __version__ = _omnivore_version.version
except ImportError:
    __version__ = "dev"

import logging
log = logging.getLogger(__name__)


# py2exe utilities
def get_py2exe_toolkit_includes(module=None, toolkit="wx"):
    """Get a list of modules that should be included in a py2exe/py2app
    distribution but because they are imported dynamically must be explicitly
    included.
    
    This list of files should be added to options keyword parameter of the
    setup call; e.g:
    
    setup(options={"py2exe": {"includes": omnivore.get_py2exe_toolkit_includes()}})
    """
    if module is None:
        path = __file__
    else:
        path = module.__file__
    import os
    basedir, module = os.path.split(os.path.dirname(path))
    #print basedir, module
    suffix = "_%s.py" % toolkit
    includes = []
    for root, dirs, files in os.walk(os.path.join(basedir, module)):
        #print root, files
        mod_root = root[len(basedir) + 1:]
        needed = [f for f in files if f.endswith(suffix)]
        for f in needed:
            mod_name = ("%s.%s" % (mod_root, f[:-3])).replace("/", ".").replace("\\", ".")
            includes.append(mod_name)
    return includes


def get_py2exe_data_files(module=None, excludes=[]):
    """Get a list of data files that should be included in a py2exe/py2app
    distribution's data files parameter.
    
    The returned list should be added to the data_files keyword parameter of the
    setup call; e.g:
    
    data_files = []
    import omnivore
    data_files.extend(omnivore.get_py2exe_data_files())
    setup(..., data_files=data_files, ...)
    """
    # added .zip files for pyface image libraries
    extensions = [".png", ".jpg", ".ico", ".zip"]

    if module is None:
        path = __file__
    else:
        path = module.__file__
    import os
    import fnmatch
    basedir, module = os.path.split(os.path.dirname(path))
    #print basedir, module
    data_files = []
    for root, dirs, files in os.walk(os.path.join(basedir, module)):
        #print root, files
        mod_root = root[len(basedir) + 1:]
        needed = []
        for f in files:
            for suffix in extensions:
                if f.endswith(suffix):
                    path = os.path.join(root, f)
                    for pattern in excludes:
                        if fnmatch.fnmatch(path, pattern):
                            path = None
                            break
                    if path:
                        needed.append(path)
                        break
        if needed:
            data_files.append((mod_root, needed))
    return data_files


def get_image_path(rel_path, module=None, file=None, up_one_level=False, excludes=[]):
    """Get the image path for static images relative to the specified module
    or file.
    
    The image path will be modified to find images in py2exe/py2app
    locations assuming that the data files have been added using the above
    get_py2exe_data_files function.
    
    Either the module or file keyword parameter may be specified to provide
    a relative module or file name.  The module may be specified either by
    reference to an imported module, or by a dotted string.  The file must be
    specified using the __file__ keyword.  If both are specified, file takes
    precedence.
    
    For example, in omnivore, the images are located in a directory "icons" in
    main omnivore directory (e.g. omnivore/icons):
        
    import omnivore
    image_path = get_image_path("icons", omnivore)
    
    will return the absolute path of "omnivore/icons".
    
    An example using the file keyword: if the current module is in the
    omnivore/framework directory, then the call to:
    
    get_image_path("icons", file=__file__)
    
    will contain the absolute path to the omnivore/framework/icons directory.
    """
    if file is None:
        if module is None:
            path = __file__
        else:
            try:
                path = module.__file__
            except AttributeError:
                # must be a string containing the module hiearchy
                path = module.replace(".", "/") + "/this-will-be-removed.py"
    else:
        path = file.replace(".", "/")
    import os
    import sys
    if up_one_level:
        path = os.path.dirname(path)
    log.debug("get_image_path: file=%s, path=%s, module=%s" % (file, path, module))
    frozen = getattr(sys, 'frozen', False)
    image_path = os.path.normpath(os.path.join(os.path.dirname(path), rel_path))
    if frozen:
        if frozen == True:
            # pyinstaller sets frozen=True and uses sys._MEIPASS
            root = sys._MEIPASS
            image_path = os.path.normpath(os.path.join(root, image_path))
        elif frozen in ('macosx_app'):
            #print "FROZEN!!! %s" % frozen
            root = os.environ['RESOURCEPATH']
            if ".zip/" in image_path:
                zippath, image_path = image_path.split(".zip/")
            image_path = os.path.normpath(os.path.join(root, image_path))
        else:
            log.error("App packager %s not yet supported for image paths!!!")
    elif not os.path.isabs(path):
        if "/" in path:
            toplevel, modulepath = path.split("/", 1)
            modulepath = os.path.dirname(modulepath)
        else:
            toplevel = path
            modulepath = ""
        path = os.path.dirname(sys.modules[toplevel].__file__)
        rel_path = os.path.join(modulepath, rel_path)
        log.debug("get_image_path: relative! path=%s toplevel=%s modulepath=%s rel_path=%s" % (path, toplevel, modulepath, rel_path))
        image_path = os.path.normpath(os.path.join(path, rel_path))
    log.debug("get_image_path: image_path=%s" % image_path)
    return image_path



"""
fs.rpcfs
========

This module provides the class 'RPCFS' to access a remote FS object over
XML-RPC.  You probably want to use this in conjunction with the 'RPCFSServer'
class from the :mod:`fs.expose.xmlrpc` module.

"""

import xmlrpclib
import socket
import base64

from fs.base import *
from fs.errors import *
from fs.path import *
from fs import iotools

from fs.filelike import StringIO

import six
from six import PY3, b


def re_raise_faults(func):
    """Decorator to re-raise XML-RPC faults as proper exceptions."""
    def wrapper(*args, **kwds):
        try:
            return func(*args, **kwds)
        except (xmlrpclib.Fault), f:
            #raise
            # Make sure it's in a form we can handle

            print f.faultString
            bits = f.faultString.split(" ")
            if bits[0] not in ["<type", "<class"]:
                raise f
            # Find the class/type object
            bits = " ".join(bits[1:]).split(">:")
            cls = bits[0]
            msg = ">:".join(bits[1:])
            cls = cls.strip('\'')
            print "-" + cls
            cls = _object_by_name(cls)
            # Re-raise using the remainder of the fault code as message
            if cls:
                if issubclass(cls, FSError):
                    raise cls('', msg=msg)
                else:
                    raise cls(msg)
            raise f
        except socket.error, e:
            raise RemoteConnectionError(str(e), details=e)
    return wrapper


def _object_by_name(name, root=None):
    """Look up an object by dotted-name notation."""
    bits = name.split(".")
    if root is None:
        try:
            obj = globals()[bits[0]]
        except KeyError:
            try:
                obj = __builtins__[bits[0]]
            except KeyError:
                obj = __import__(bits[0], globals())
    else:
        obj = getattr(root, bits[0])
    if len(bits) > 1:
        return _object_by_name(".".join(bits[1:]), obj)
    else:
        return obj


class ReRaiseFaults:
    """XML-RPC proxy wrapper that re-raises Faults as proper Exceptions."""

    def __init__(self, obj):
        self._obj = obj

    def __getattr__(self, attr):
        val = getattr(self._obj, attr)
        if callable(val):
            val = re_raise_faults(val)
            self.__dict__[attr] = val
        return val


class RPCFS(FS):
    """Access a filesystem exposed via XML-RPC.

    This class provides the client-side logic for accessing a remote FS
    object, and is dual to the RPCFSServer class defined in fs.expose.xmlrpc.

    Example::

        fs = RPCFS("http://my.server.com/filesystem/location/")

    """

    _meta = {'thread_safe' : True,
             'virtual': False,
             'network' : True,
              }

    def __init__(self, uri, transport=None):
        """Constructor for RPCFS objects.

        The only required argument is the URI of the server to connect
        to.  This will be passed to the underlying XML-RPC server proxy
        object, along with the 'transport' argument if it is provided.

        :param uri: address of the server

        """
        super(RPCFS, self).__init__(thread_synchronize=True)
        self.uri = uri
        self._transport = transport
        self.proxy = self._make_proxy()
        self.isdir('/')

    @synchronize
    def _make_proxy(self):
        kwds = dict(allow_none=True, use_datetime=True)

        if self._transport is not None:
            proxy = xmlrpclib.ServerProxy(self.uri, self._transport, **kwds)
        else:
            proxy = xmlrpclib.ServerProxy(self.uri, **kwds)

        return ReRaiseFaults(proxy)

    def __str__(self):
        return '<RPCFS: %s>' % (self.uri,)

    def __repr__(self):
        return '<RPCFS: %s>' % (self.uri,)

    @synchronize
    def __getstate__(self):
        state = super(RPCFS, self).__getstate__()
        try:
            del state['proxy']
        except KeyError:
            pass
        return state

    def __setstate__(self, state):
        super(RPCFS, self).__setstate__(state)
        self.proxy = self._make_proxy()

    def encode_path(self, path):
        """Encode a filesystem path for sending over the wire.

        Unfortunately XMLRPC only supports ASCII strings, so this method
        must return something that can be represented in ASCII.  The default
        is base64-encoded UTF8.
        """
        return six.text_type(base64.b64encode(path.encode("utf8")), 'ascii')

    def decode_path(self, path):
        """Decode paths arriving over the wire."""
        return six.text_type(base64.b64decode(path.encode('ascii')), 'utf8')

    @synchronize
    def getmeta(self, meta_name, default=NoDefaultMeta):
        if default is NoDefaultMeta:
            meta = self.proxy.getmeta(meta_name)
        else:
            meta = self.proxy.getmeta_default(meta_name, default)
        if isinstance(meta, basestring):
            #  To allow transport of meta with invalid xml chars (like null)
            meta = self.encode_path(meta)
        return meta

    @synchronize
    def hasmeta(self, meta_name):
        return self.proxy.hasmeta(meta_name)

    @synchronize
    @iotools.filelike_to_stream
    def open(self, path, mode='r', buffering=-1, encoding=None, errors=None, newline=None, line_buffering=False, **kwargs):
        # TODO: chunked transport of large files
        epath = self.encode_path(path)
        if "w" in mode:
            self.proxy.set_contents(epath, xmlrpclib.Binary(b("")))
        if "r" in mode or "a" in mode or "+" in mode:
            try:
                data = self.proxy.get_contents(epath, "rb").data
            except IOError:
                if "w" not in mode and "a" not in mode:
                    raise ResourceNotFoundError(path)
                if not self.isdir(dirname(path)):
                    raise ParentDirectoryMissingError(path)
                self.proxy.set_contents(path, xmlrpclib.Binary(b("")))
        else:
            data = b("")
        f = StringIO(data)
        if "a" not in mode:
            f.seek(0, 0)
        else:
            f.seek(0, 2)
        oldflush = f.flush
        oldclose = f.close
        oldtruncate = f.truncate

        def newflush():
            self._lock.acquire()
            try:
                oldflush()
                self.proxy.set_contents(epath, xmlrpclib.Binary(f.getvalue()))
            finally:
                self._lock.release()

        def newclose():
            self._lock.acquire()
            try:
                f.flush()
                oldclose()
            finally:
                self._lock.release()

        def newtruncate(size=None):
            self._lock.acquire()
            try:
                oldtruncate(size)
                f.flush()
            finally:
                self._lock.release()

        f.flush = newflush
        f.close = newclose
        f.truncate = newtruncate
        return f

    @synchronize
    def exists(self, path):
        path = self.encode_path(path)
        return self.proxy.exists(path)

    @synchronize
    def isdir(self, path):
        path = self.encode_path(path)
        return self.proxy.isdir(path)

    @synchronize
    def isfile(self, path):
        path = self.encode_path(path)
        return self.proxy.isfile(path)

    @synchronize
    def listdir(self, path="./", wildcard=None, full=False, absolute=False, dirs_only=False, files_only=False):
        enc_path = self.encode_path(path)
        if not callable(wildcard):
            entries = self.proxy.listdir(enc_path,
                                         wildcard,
                                         full,
                                         absolute,
                                         dirs_only,
                                         files_only)
            entries = [self.decode_path(e) for e in entries]
        else:
            entries = self.proxy.listdir(enc_path,
                                         None,
                                         False,
                                         False,
                                         dirs_only,
                                         files_only)
            entries = [self.decode_path(e) for e in entries]
            entries = [e for e in entries if wildcard(e)]
            if full:
                entries = [relpath(pathjoin(path, e)) for e in entries]
            elif absolute:
                entries = [abspath(pathjoin(path, e)) for e in entries]
        return entries

    @synchronize
    def makedir(self, path, recursive=False, allow_recreate=False):
        path = self.encode_path(path)
        return self.proxy.makedir(path, recursive, allow_recreate)

    @synchronize
    def remove(self, path):
        path = self.encode_path(path)
        return self.proxy.remove(path)

    @synchronize
    def removedir(self, path, recursive=False, force=False):
        path = self.encode_path(path)
        return self.proxy.removedir(path, recursive, force)

    @synchronize
    def rename(self, src, dst):
        src = self.encode_path(src)
        dst = self.encode_path(dst)
        return self.proxy.rename(src, dst)

    @synchronize
    def settimes(self, path, accessed_time, modified_time):
        path = self.encode_path(path)
        return self.proxy.settimes(path, accessed_time, modified_time)

    @synchronize
    def getinfo(self, path):
        path = self.encode_path(path)
        info = self.proxy.getinfo(path)
        return info

    @synchronize
    def desc(self, path):
        path = self.encode_path(path)
        return self.proxy.desc(path)

    @synchronize
    def getxattr(self, path, attr, default=None):
        path = self.encode_path(path)
        attr = self.encode_path(attr)
        return self.fs.getxattr(path, attr, default)

    @synchronize
    def setxattr(self, path, attr, value):
        path = self.encode_path(path)
        attr = self.encode_path(attr)
        return self.fs.setxattr(path, attr, value)

    @synchronize
    def delxattr(self, path, attr):
        path = self.encode_path(path)
        attr = self.encode_path(attr)
        return self.fs.delxattr(path, attr)

    @synchronize
    def listxattrs(self, path):
        path = self.encode_path(path)
        return [self.decode_path(a) for a in self.fs.listxattrs(path)]

    @synchronize
    def copy(self, src, dst, overwrite=False, chunk_size=16384):
        src = self.encode_path(src)
        dst = self.encode_path(dst)
        return self.proxy.copy(src, dst, overwrite, chunk_size)

    @synchronize
    def move(self, src, dst, overwrite=False, chunk_size=16384):
        src = self.encode_path(src)
        dst = self.encode_path(dst)
        return self.proxy.move(src, dst, overwrite, chunk_size)

    @synchronize
    def movedir(self, src, dst, overwrite=False, ignore_errors=False, chunk_size=16384):
        src = self.encode_path(src)
        dst = self.encode_path(dst)
        return self.proxy.movedir(src, dst, overwrite, ignore_errors, chunk_size)

    @synchronize
    def copydir(self, src, dst, overwrite=False, ignore_errors=False, chunk_size=16384):
        src = self.encode_path(src)
        dst = self.encode_path(dst)
        return self.proxy.copydir(src, dst, overwrite, ignore_errors, chunk_size)



#------------------------------------------------------------------------------
# Copyright (c) 2005, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought logger package component>
#------------------------------------------------------------------------------

# Standard library imports
from datetime import datetime
import logging

# Enthought library imports.
from pyface.api import ImageResource, clipboard
from pyface.workbench.api import TraitsUIView
from traits.api import Button, Instance, List, Property, Str, \
    cached_property, on_trait_change
from traitsui.api import View, Group, Item, CodeEditor, \
    TabularEditor, spring
from traitsui.tabular_adapter import TabularAdapter

# Local imports
from apptools.logger.agent.quality_agent_view import QualityAgentView
from apptools.logger.plugin import view
from apptools.logger.plugin.logger_service import LoggerService

# Constants
_IMAGE_MAP = { logging.DEBUG: ImageResource('debug'),
               logging.INFO: ImageResource('info'),
               logging.WARNING: ImageResource('warning'),
               logging.ERROR: ImageResource('error'),
               logging.CRITICAL: ImageResource('crit_error') }


class LogRecordAdapter(TabularAdapter):
    """ A TabularEditor adapter for logging.LogRecord objects.
    """

    columns = [ ('Level', 'level'), ('Date', 'date'), ('Time', 'time'),
                ('Message', 'message') ]
    column_widths = [ 80, 100, 120, -1 ]

    level_image = Property
    level_text = Property(Str)
    date_text = Property(Str)
    time_text = Property(Str)
    message_text = Property(Str)

    def get_width(self, object, trait, column):
        return self.column_widths[column]

    def _get_level_image(self):
        return _IMAGE_MAP[self.item.levelno]

    def _get_level_text(self):
        return self.item.levelname.capitalize()

    def _get_date_text(self):
        dt = datetime.fromtimestamp(self.item.created)
        return dt.date().isoformat()

    def _get_time_text(self):
        dt = datetime.fromtimestamp(self.item.created)
        return dt.time().isoformat()

    def _get_message_text(self):
        # Just display the first line of multiline messages, like stacktraces.
        msg = self.item.getMessage()
        msgs = msg.strip().split('\n')
        if len(msgs) > 1:
            suffix = '... [double click for details]'
        else:
            suffix = ''
        abbrev_msg = msgs[0] + suffix
        return abbrev_msg


class LoggerView(TraitsUIView):
    """ The Workbench View showing the list of log items.
    """

    id = Str('apptools.logger.plugin.view.logger_view.LoggerView')
    name = Str('Logger')
    service = Instance(LoggerService)

    log_records = List(Instance(logging.LogRecord))
    formatted_records = Property(Str, depends_on='log_records')

    activated = Instance(logging.LogRecord)
    activated_text = Property(Str, depends_on='activated')
    reset_button = Button("Reset Logs")
    show_button = Button("Complete Text Log")
    copy_button = Button("Copy Log to Clipboard")


    code_editor = CodeEditor(lexer='null',
                             show_line_numbers=False)
    log_records_editor = TabularEditor(adapter=LogRecordAdapter(),
                                       editable=False,
                                       activated='activated')
    trait_view = View(Group(Item('log_records',
                                 editor=log_records_editor),
                            Group(Item('reset_button'),
                                  spring,
                                  Item('show_button'),
                                  Item('copy_button'),
                                  orientation='horizontal',
                                  show_labels=False),
                            show_labels=False))

    ###########################################################################
    # LogQueueHandler view interface
    ###########################################################################

    def update(self, force=False):
        """ Update 'log_records' if our handler has new records or 'force' is
            set.
        """
        service = self.service
        if service.handler.has_new_records() or force:
            log_records = [ rec for rec in service.handler.get()
                            if rec.levelno >= service.preferences.level_ ]
            log_records.reverse()
            self.log_records = log_records

    ###########################################################################
    # Private interface
    ###########################################################################

    @on_trait_change('service.preferences.level_')
    def _update_log_records(self):
        self.service.handler._view = self
        self.update(force=True)

    def _reset_button_fired(self):
        self.service.handler.reset()
        self.log_records = []

    def _show_button_fired(self):
        self.edit_traits(view=View(Item('formatted_records',
                                        editor=self.code_editor,
                                        style='readonly',
                                        show_label=False),
                                   width=800, height=600, resizable=True,
                                   buttons=[ 'OK' ],
                                   title='Complete Text Log'))

    def _copy_button_fired(self):
        clipboard.text_data = self.formatted_records

    @cached_property
    def _get_formatted_records(self):
        return '\n'.join([ self.service.handler.formatter.format(record)
                           for record in self.log_records ])

    def _activated_changed(self):
        if self.activated is None:
            return
        msg = self.activated.getMessage()
        if self.service.preferences.enable_agent:
            dialog = QualityAgentView(msg=msg, service=self.service)
            dialog.open()
        else:
            self.edit_traits(view=View(Item('activated_text',
                                            editor=self.code_editor,
                                            style='readonly',
                                            show_label=False),
                                       width=800, height=600, resizable=True,
                                       buttons=[ 'OK' ],
                                       title='Log Message Detail'))

    @cached_property
    def _get_activated_text(self):
        if self.activated is None:
            return ''
        else:
            return self.activated.getMessage()

#### EOF ######################################################################



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/21/2004
#
#------------------------------------------------------------------------------

""" Defines the various editors for multi-selection enumerations, for the
wxPython user interface toolkit.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import logging

import wx

from traits.api \
    import List, Str, TraitError

# FIXME: ToolkitEditorFactory is a proxy class defined here just for backward
# compatibility. The class has been moved to the
# traitsui.editors.check_list_editor file.
from traitsui.editors.check_list_editor \
    import ToolkitEditorFactory

from editor_factory \
    import TextEditor as BaseTextEditor

from editor \
    import EditorWithList

from helper \
    import TraitsUIPanel
from functools import reduce

logger = logging.getLogger(__name__)


# default formatting function (would import from string, but not in Python 3)
capitalize = lambda s: s.capitalize()


#-------------------------------------------------------------------------
#  'SimpleEditor' class:
#-------------------------------------------------------------------------

class SimpleEditor(EditorWithList):
    """ Simple style of editor for checklists, which displays a combo box.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # Checklist item names
    names = List(Str)

    # Checklist item values
    values = List

    #-------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #-------------------------------------------------------------------------

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self.create_control(parent)
        super(SimpleEditor, self).init(parent)
        self.set_tooltip()

    #-------------------------------------------------------------------------
    #  Creates the initial editor control:
    #-------------------------------------------------------------------------

    def create_control(self, parent):
        """ Creates the initial editor control.
        """
        self.control = wx.Choice(parent, -1,
                                 wx.Point(0, 0), wx.Size(100, 20), [])
        wx.EVT_CHOICE(parent, self.control.GetId(), self.update_object)

    #-------------------------------------------------------------------------
    #  Handles the list of legal check list values being updated:
    #-------------------------------------------------------------------------

    def list_updated(self, values):
        """ Handles updates to the list of legal checklist values.
        """
        sv = self.string_value
        if (len(values) > 0) and isinstance(values[0], basestring):
            values = [(x, sv(x, capitalize)) for x in values]
        self.values = valid_values = [x[0] for x in values]
        self.names = [x[1] for x in values]

        # Make sure the current value is still legal:
        modified = False
        cur_value = parse_value(self.value)
        for i in range(len(cur_value) - 1, -1, -1):
            if cur_value[i] not in valid_values:
                try:
                    del cur_value[i]
                    modified = True
                except TypeError as e:
                    logger.warn('Unable to remove non-current value [%s] from '
                                'values %s', cur_value[i], values)
        if modified:
            if isinstance(self.value, basestring):
                cur_value = ','.join(cur_value)
            self.value = cur_value

        self.rebuild_editor()

    #-------------------------------------------------------------------------
    #  Rebuilds the editor after its definition is modified:
    #-------------------------------------------------------------------------

    def rebuild_editor(self):
        """ Rebuilds the editor after its definition is modified.
        """
        control = self.control
        control.Clear()
        for name in self.names:
            control.Append(name)

        self.update_editor()

    #-------------------------------------------------------------------------
    #  Handles the user selecting a new value from the combo box:
    #-------------------------------------------------------------------------

    def update_object(self, event):
        """ Handles the user selecting a new value from the combo box.
        """
        value = self.values[self.names.index(event.GetString())]
        if not isinstance(self.value, str):
            value = [value]

        self.value = value

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        try:
            self.control.SetSelection(
                self.values.index(parse_value(self.value)[0]))
        except:
            pass

#-------------------------------------------------------------------------
#  'CustomEditor' class:
#-------------------------------------------------------------------------


class CustomEditor(SimpleEditor):
    """ Custom style of editor for checklists, which displays a set of check
        boxes.
    """

    #-------------------------------------------------------------------------
    #  Creates the initial editor control:
    #-------------------------------------------------------------------------

    def create_control(self, parent):
        """ Creates the initial editor control.
        """
        # Create a panel to hold all of the check boxes
        self.control = panel = TraitsUIPanel(parent, -1)

    #-------------------------------------------------------------------------
    #  Rebuilds the editor after its definition is modified:
    #-------------------------------------------------------------------------

    def rebuild_editor(self):
        """ Rebuilds the editor after its definition is modified.
        """
        panel = self.control
        panel.SetSizer(None)
        panel.DestroyChildren()

        cur_value = parse_value(self.value)

        # Create a sizer to manage the radio buttons:
        labels = self.names
        values = self.values
        n = len(labels)
        cols = self.factory.cols
        rows = (n + cols - 1) / cols
        incr = [n / cols] * cols
        rem = n % cols

        for i in range(cols):
            incr[i] += (rem > i)
        incr[-1] = -(reduce(lambda x, y: x + y, incr[:-1], 0) - 1)

        if cols > 1:
            sizer = wx.GridSizer(0, cols, 2, 4)
        else:
            sizer = wx.BoxSizer(wx.VERTICAL)

        # Add the set of all possible choices:
        index = 0
        for i in range(rows):
            for j in range(cols):
                if n > 0:
                    label = labels[index]
                    control = wx.CheckBox(panel, -1, label)
                    control.value = value = values[index]
                    control.SetValue(value in cur_value)
                    wx.EVT_CHECKBOX(panel, control.GetId(), self.update_object)
                    index += incr[j]
                    n -= 1
                else:
                    control = wx.CheckBox(panel, -1, '')
                    control.Show(False)

                sizer.Add(control, 0, wx.NORTH, 5)

        # Lay out the controls:
        panel.SetSizerAndFit(sizer)

        # FIXME: There are cases where one of the parent panel's of the check
        # list editor has a fixed 'min size' which prevents the check list
        # editor from expanding correctly, so we currently are making sure
        # that all of the parent panels do not have a fixed min size before
        # doing the layout/refresh:
        parent = panel.GetParent()
        while isinstance(parent, wx.Panel):
            parent.SetMinSize(wx.Size(-1, -1))
            panel = parent
            parent = parent.GetParent()

        panel.Layout()
        panel.Refresh()

    #-------------------------------------------------------------------------
    #  Handles the user clicking one of the 'custom' check boxes:
    #-------------------------------------------------------------------------

    def update_object(self, event):
        """ Handles the user clicking one of the custom check boxes.
        """
        control = event.GetEventObject()
        cur_value = parse_value(self.value)
        if control.GetValue():
            cur_value.append(control.value)
        else:
            cur_value.remove(control.value)
        if isinstance(self.value, basestring):
            cur_value = ','.join(cur_value)
        self.value = cur_value

    #-------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #-------------------------------------------------------------------------

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        new_values = parse_value(self.value)
        for control in self.control.GetChildren():
            if control.IsShown():
                control.SetValue(control.value in new_values)

#-------------------------------------------------------------------------
#  'TextEditor' class:
#-------------------------------------------------------------------------


class TextEditor(BaseTextEditor):
    """ Text style of editor for checklists, which displays a text field.
    """

    #-------------------------------------------------------------------------
    #  Handles the user changing the contents of the edit control:
    #-------------------------------------------------------------------------

    def update_object(self, event):
        """ Handles the user changing the contents of the edit control.
        """
        try:
            value = self.control.GetValue()
            value = eval(value)
        except:
            pass
        try:
            self.value = value
        except TraitError as excp:
            pass

#-------------------------------------------------------------------------
#  Parse a value into a list:
#-------------------------------------------------------------------------


def parse_value(value):
    """ Parses a value into a list.
    """
    if value is None:
        return []

    if not isinstance(value, str):
        return value[:]

    return [x.strip() for x in value.split(',')]

### EOF ##################################################################



from abstract_adapter_factory import AbstractAdapterFactory
from abstract_factory import AbstractFactory
from abstract_type_system import AbstractTypeSystem
from adaptable import Adaptable
from adapter import Adapter
from adapter_factory import AdapterFactory
from adapter_manager import AdapterManager
from factory import Factory
from hook import add_pre, add_post, remove_pre, remove_post
from python_type_system import PythonObject, PythonTypeSystem
from type_manager import TypeManager



import os

# Major package imports.
import numpy as np
from fs.opener import opener
import wx
import fs

# Enthought library imports.
from traits.api import on_trait_change, Any, Bool, Int, Unicode, Property, Dict, List
from pyface.api import YES, NO
from pyface.tasks.api import Editor
from pyface.action.api import ActionEvent

from omnivore import __version__
from omnivore.utils.command import HistoryList, StatusFlags
from omnivore.utils.sortutil import collapse_overlapping_ranges, invert_ranges, ranges_to_indexes
from omnivore.utils.file_guess import FileGuess
from omnivore.framework.document import DocumentError

import logging
log = logging.getLogger(__name__)


class FrameworkEditor(Editor):
    """The pyface editor template for the omnivore framework
    
    The abstract methods 
    """

    #### 'IProjectEditor' interface ############################################

    document = Any(None)

    name = Property(Unicode, depends_on='document')

    tooltip = Property(Unicode, depends_on='document')

    can_save = Bool(True)

    can_undo = Bool(False)  # has to be a property of Editor because EditorActions need to refer to this trait

    undo_label = Unicode("Undo")

    can_redo = Bool(False)

    redo_label = Unicode("Redo")

    printable = Bool(False)

    imageable = Bool(False)

    can_cut = Bool(False)

    can_copy = Bool(False)

    # can_paste is set by the idle handler of the FrameworkApplication for the
    # active window
    can_paste = Bool(False)

    # Cursor index points to positions between bytes, so zero is before the
    # first byte and the max index is the number of bytes, which points to
    # after the last byte

    cursor_index = Int(0)

    cursor_history = Any

    # Anchor indexes behave like cursor positions: they indicate positions
    # between bytes
    anchor_start_index = Int(0)

    anchor_initial_start_index = Int(0)

    anchor_initial_end_index = Int(0)

    anchor_end_index = Int(0)

    selected_ranges = List([])

    last_search_settings = Dict()

    mouse_mode_factory = Any

    baseline_present = Bool

    diff_highlight = Bool

    _metadata_dirty = Bool(transient=True)

    #### trait default values

    def _document_default(self):
        return self.task.window.application.document_class()

    def _cursor_history_default(self):
        return HistoryList()

    def _last_search_settings_default(self):
        return {
            "find": "",
            "replace": "",
            "match_case": False,
            "allow_inverse": False,
            "regex": False,
            "algorithm": "",
            }

    def _selected_ranges_default(self):
        return [(0, 0)]

    #### property getters

    def _get_name(self):
        return self.document.name

    def _get_tooltip(self):
        return self.document.metadata.uri

    ###########################################################################
    # 'FrameworkEditor' interface.
    ###########################################################################

    def create(self, parent):
        """ Creates the toolkit-specific control for the widget.
        """
        raise NotImplementedError

    def load(self, source=None):
        """ Loads the contents of the editor.
        """
        if source is None:
            doc = self.task.window.application.document_class()
        elif hasattr(source, 'document_id'):
            self.init_extra_metadata(source)
            self.view_document(source)
        else:
            doc = self.task.window.application.guess_document(source)
            self.init_extra_metadata(doc)
            self.view_document(doc)

    def init_extra_metadata(self, doc):
        """ Hook to load any extra metadata for the given document
        """
        e = doc.get_metadata_for(self.task)
        self.from_metadata_dict(e)
        self.metadata_dirty = False

    def get_extra_metadata_header(self):
        return "# omnivore %s extra_metadata=v1\n" % __version__

    def from_metadata_dict(self, e):
        """ Set up additional object attributes from the dict argument
        """
        pass

    def to_metadata_dict(self, metadata_dict, document):
        """ Store any persistent object attributes in a dictionary that
        will be serialized and later loaded in from_metadata_dict
        """
        pass

    def load_baseline(self, uri, doc=None):
        if doc is None:
            doc = self.document
        try:
            doc.load_baseline(uri, confirm_callback=self.task.confirm)
        except DocumentError, e:
            self.window.error("Failed opening baseline document file\n\n%s\n\nError: %s" % (uri, str(e)), "Baseline Document Loading Error")
            return
        if doc == self.document:
            self.baseline_present = doc.has_baseline
            self.diff_highlight = self.baseline_present

    def use_self_as_baseline(self, doc=None):
        if doc is None:
            doc = self.document
        bytes = np.copy(doc.bytes)
        doc.init_baseline(doc.metadata, bytes)
        if doc == self.document:
            self.baseline_present = doc.has_baseline
            self.diff_highlight = self.baseline_present

    def view_document(self, doc, old_editor=None):
        """ Change the view to the specified document
        """
        doc.last_task_id = self.task.id
        self.document = self.task.window.application.add_document(doc)
        self.rebuild_document_properties()
        if old_editor is not None:
            self.copy_view_properties(old_editor)
        else:
            self.init_view_properties()
        self.rebuild_ui()
        self.document.undo_stack_changed = True
        self.task.document_changed = self.document

    def rebuild_document_properties(self):
        """ Recreate any editor attributes for the new document
        """
        self.baseline_present = self.document.has_baseline

    def init_view_properties(self):
        """ Set up editor properties when loading a new file
        """
        pass

    def copy_view_properties(self, old_editor):
        """ Copy editor properties to the new view
        """
        pass

    def process_preference_change(self, prefs):
        """ Update any values dependent on application preferences
        """
        pass

    @property
    def document_length(self):
        return len(self.document)

    def save(self, uri=None, saver=None, document=None):
        """ Saves a document.

        It can be used 3 ways:

        1) save the contents of the editor
        2) save the specified document using the save format of this editor.
        3) save a document using the specified save conversion routine
        """
        if document is None:
            document = self.document

            # only check validity if saving the current document
            if not self.is_valid_for_save():
                document.undo_stack_changed = True
                return

        if uri is None:
            uri = document.uri

        try:
            document.save_to_uri(uri, self, saver)
            document.undo_stack.set_save_point()

            # force an update to the document name as the URI may have changed
            document.uri = uri

            # force update: just got saved, can't be read only!
            document.read_only = False

            # Update tab name.  Note that dirty must be changed in order for
            # the trait to be updated, so force a change if needed.  Also,
            # update the URI first because trait callbacks happen immediately
            # and because properties are used for the editor name, no trait
            # event gets called on updating the metadata URI.
            if document == self.document:
                if not self.dirty:
                    self.dirty = True
                self.metadata_dirty = False

            document.undo_stack_changed = True
            saved = True

            # refresh window name in case filename has changed
            self.task._active_editor_tab_change(None)
        except Exception, e:
            import traceback
            stack = traceback.format_exc()
            log.error("%s:\n%s" % (uri, stack))
            self.window.error("Error trying to save:\n\n%s\n\n%s" % (uri, str(e)), "File Save Error")
            saved = False
        return saved

    def is_valid_for_save(self):
        """Hook for subclasses to implement a validity check before saving
        file
        """
        return True

    # Segment saver interface for menu item display
    export_data_name = "Any"
    export_extensions = [".*"]

    def encode_data(self, document, editor):
        """Document saver interface: take a document and produce a byte
        representation to save to disk.
        """
        data = document.bytes.tostring()
        return data

    @property
    def metadata_dirty(self):
        return self._metadata_dirty

    @metadata_dirty.setter
    def metadata_dirty(self, val):
        self._metadata_dirty = bool(val)
        self.dirty = self.document.dirty or self._metadata_dirty
        self.can_save = not self.document.read_only and self.dirty

    def undo(self):
        """ Undoes the last action
        """
        raise NotImplementedError

    def redo(self):
        """ Re-performs the last undone action
        """
        raise NotImplementedError

    def cut(self):
        """ Copies the current selection to the clipboard and removes the selection
        """
        raise NotImplementedError

    def copy(self):
        """ Copies the current selection to the clipboard
        """
        data_obj = self.create_clipboard_data_object()
        if data_obj is None:
            return False
        return self.set_clipboard_object(data_obj)

    def set_clipboard_object(self, data_obj):
        if wx.TheClipboard.Open():
            wx.TheClipboard.SetData(data_obj)
            self.show_data_object_stats(data_obj)
            wx.TheClipboard.Close()
            return True
        else:
            self.window.error("Unable to open clipboard", "Clipboard Error")
            return False

    def get_data_object_by_format(self, data_obj, fmt):
        # First try a composite object, then simple: have to handle both
        # cases
        try:
            d = data_obj.GetObject(fmt)
        except AttributeError:
            d = data_obj
        return d

    def show_data_object_stats(self, data_obj, copy=True):
        if wx.DF_TEXT in data_obj.GetAllFormats():
            fmt = wx.DataFormat(wx.DF_TEXT)
        elif wx.DF_UNICODETEXT in data_obj.GetAllFormats():  # for windows
            fmt = wx.DataFormat(wx.DF_UNICODETEXT)
        else:
            fmt = None

        if fmt is not None:
            d = self.get_data_object_by_format(data_obj, fmt)
            size = d.GetDataSize()
            self.task.status_bar.message = "%s %d text characters" % ("Copied" if copy else "Pasted", size)

    def paste(self, cmd_cls=None):
        """ Pastes the current clipboard at the current insertion point or over
        the current selection
        """
        data_obj = self.get_paste_data_object()
        if data_obj is not None:
            self.process_paste_data_object(data_obj, cmd_cls)
            self.show_data_object_stats(data_obj, False)
        else:
            self.window.error("Unsupported data format", "Paste Error")

    def get_paste_data_object(self):
        data_objs = self.supported_clipboard_data_objects

        if wx.TheClipboard.Open():
            for data_obj in data_objs:
                success = wx.TheClipboard.GetData(data_obj)
                if success:
                    break
            wx.TheClipboard.Close()
        else:
            self.window.error("Unable to open clipboard", "Clipboard Error")
            success = False
        if success:
            return data_obj
        return None

    def process_paste_data_object(self, data_obj, cmd_cls=None):
        pass  # Override in subclass

    # must be a class attribute because for checking clipboard data formats of
    # custom objects, data formats must be singletons
    supported_clipboard_data_objects = [wx.TextDataObject()]

    def create_clipboard_data_object(self):
        return wx.TextDataObject("Omnivore!")

    def print_preview(self):
        raise NotImplementedError

    def print_page(self):
        raise NotImplementedError

    def save_as_pdf(self, path=None):
        raise NotImplementedError

    def save_as_image(self, path):
        """ Saves the contents of the editor in a maproom project file
        """
        valid = {
            '.png': wx.BITMAP_TYPE_PNG,
            '.tif': wx.BITMAP_TYPE_TIFF,
            '.tiff': wx.BITMAP_TYPE_TIFF,
            '.jpg': wx.BITMAP_TYPE_JPEG,
            '.jpeg': wx.BITMAP_TYPE_JPEG,
            }
        _, ext = os.path.splitext(path)
        if ext not in valid:
            path += ".png"
            t = wx.BITMAP_TYPE_PNG
        else:
            t = valid[ext]

        raw_data = self.get_numpy_image()
        h, w, depth = raw_data.shape
        bitmap = wx.BitmapFromBuffer(width=w, height=h, dataBuffer=raw_data)
        image = wx.ImageFromBitmap(bitmap)
        image.SaveFile(path, t)

    def get_numpy_image(self):
        """Get a numpy array in RGB format to be saved to a file
        """
        raise NotImplementedError

    def update_history(self):
        """Hook to update any undo history list and/or save history log to a file
        """
        pass

    def made_current_active_editor(self):
        pass

    def rebuild_ui(self):
        """Called when each pane should be rebuilt from the (possibly new)
        document, or when the document formatting or structure has changed.
        """
        pass

    def refresh_panes(self):
        """Called when the panes should be repainted.
        
        Typically this is called when the contents of the document have changed
        but the document formatting or structure hasn't changed.
        """
        pass

    def reconfigure_panes(self):
        """Called when the panes should be updated after a possible document
        formatting or structure change.
        """
        pass

    def ensure_visible(self, start, end):
        """Make sure the current range of indexes is shown
        """
        pass

    def select_all(self, refresh=True):
        """ Selects the entire document
        """
        self.anchor_start_index = self.anchor_initial_start_index = 0
        self.anchor_end_index = self.anchor_initial_end_index = self.document_length
        self.selected_ranges = [(self.anchor_start_index, self.anchor_end_index)]
        self.can_copy = (self.anchor_start_index != self.anchor_end_index)
        self.highlight_selected_ranges()
        if refresh:
            self.refresh_panes()

    def select_none(self, refresh=True):
        """ Clears any selection in the document
        """
        self.anchor_start_index = self.anchor_initial_start_index = self.anchor_end_index = self.anchor_initial_end_index = self.cursor_index
        self.can_copy = False
        self.selected_ranges = [(self.cursor_index, self.cursor_index)]
        self.highlight_selected_ranges()
        if refresh:
            self.refresh_panes()

    def select_none_if_selection(self):
        if self.can_copy:
            self.select_none()

    def select_ranges(self, ranges, refresh=True):
        """ Selects the specified ranges
        """
        self.selected_ranges = ranges
        try:
            start, end = self.selected_ranges[-1]
        except IndexError:
            start, end = 0, 0
        self.anchor_start_index = self.anchor_initial_start_index = start
        self.anchor_end_index = self.anchor_initial_end_index = end
        self.can_copy = (self.anchor_start_index != self.anchor_end_index)
        self.highlight_selected_ranges()
        if refresh:
            self.refresh_panes()

    def select_invert(self, refresh=True):
        """ Selects the entire document
        """
        ranges = self.invert_selection_ranges(self.selected_ranges)
        self.select_ranges(ranges, refresh)

    def select_range(self, start, end, add=False, extend=False):
        """ Adjust the current selection to the new start and end indexes
        """
        if extend:
            self.selected_ranges[-1] = (start, end)
        elif add:
            self.selected_ranges.append((start, end))
        else:
            self.selected_ranges = [(start, end)]
        self.anchor_start_index = start
        self.anchor_end_index = end
        self.highlight_selected_ranges()

    def highlight_selected_ranges(self):
        pass

    def get_optimized_selected_ranges(self):
        """ Get the list of monotonically increasing, non-overlapping selected
        ranges
        """
        return collapse_overlapping_ranges(self.selected_ranges)

    def get_selected_ranges_and_indexes(self):
        opt = self.get_optimized_selected_ranges()
        return opt, ranges_to_indexes(opt)

    def invert_selection_ranges(self, ranges):
        return invert_ranges(ranges, self.document_length)

    def set_cursor(self, index, refresh=True):
        max_index = self.document_length - 1
        if index < 0:
            index = 0
        elif index > max_index:
            index = max_index
        self.cursor_index = index
        self.select_none(False)
        if refresh:
            self.refresh_panes()

        return index

    def update_cursor_history(self):
        state = self.get_cursor_state()
        last = self.cursor_history.get_undo_command()
        if last is None or last != state:
            cmd = self.cursor_history.get_redo_command()
            if cmd is None or cmd != state:
                self.cursor_history.add_command(state)

    def get_cursor_state(self):
        return self.cursor_index

    def undo_cursor_history(self):
        if not self.cursor_history.can_redo():
            # at the end of the history list, the last item will be the current position, so skip it
            _ = self.cursor_history.prev_command()
        cmd = self.cursor_history.prev_command()
        if cmd is None:
            return
        self.restore_cursor_state(cmd)

    def redo_cursor_history(self):
        if not self.cursor_history.can_undo():
            # at the start of the history list, the last item will be the current position, so skip it
            _ = self.cursor_history.next_command()
        cmd = self.cursor_history.next_command()
        if cmd is None:
            return
        self.restore_cursor_state(cmd)

    def restore_cursor_state(self, state):
        self.set_cursor(state)

    def mark_index_range_changed(self, index_range):
        """Hook for subclasses to be informed when bytes within the specified
        index range have changed.
        """
        pass

    def rebuild_display_objects(self):
        """Hook for subclasses to get notified when the document is changed and
        to rebuild any objects necessary for display before the call to
        refresh_panes."""
        pass

    def invalidate_search(self):
        """Hook for subclasses to get notified if the document is changed and
        any search params should be cleared."""
        pass

    def compare_to_baseline(self):
        """Hook for subclasses to update any comparisons to the baseline data."""
        pass

    def update_mouse_mode(self, mouse_handler=None):
        """Hook for subclasses to process the change to a new mouse mode 
        """
        pass

    def perform_idle(self):
        """Hook for subclasses to do some (small!) processing during UI idle
        time.
        """
        pass

    def popup_visible(self):
        """Hook for subclass to state whether a popup is being shown
        """
        return False

    def clear_popup(self):
        """Hook for subclass to remove the topmost popup
        """
        pass

    # Command processor

    def update_undo_redo(self):
        command = self.document.undo_stack.get_undo_command()
        if command is None:
            self.undo_label = "Undo"
            self.can_undo = False
        else:
            text = str(command).replace("&", "&&")
            self.undo_label = "Undo: %s" % text
            self.can_undo = True

        command = self.document.undo_stack.get_redo_command()
        if command is None:
            self.redo_label = "Redo"
            self.can_redo = False
        else:
            text = str(command).replace("&", "&&")
            self.redo_label = "Redo: %s" % text
            self.can_redo = True
        self.dirty = self.document.dirty or self._metadata_dirty
        self.can_save = not self.document.read_only and self.dirty

    def undo(self):
        undo = self.document.undo_stack.undo(self)
        self.process_flags(undo.flags)
        self.document.undo_stack_changed = True
        self.undo_post_hook()

    def undo_post_hook(self):
        pass

    def redo(self):
        undo = self.document.undo_stack.redo(self)
        self.process_flags(undo.flags)
        self.document.undo_stack_changed = True
        self.redo_post_hook()

    def redo_post_hook(self):
        pass

    def end_batch(self):
        self.document.undo_stack.end_batch()
        self.document.undo_stack_changed = True

    def process_command(self, command, batch=None):
        """Process a single command and immediately update the UI to reflect
        the results of the command.
        """
        f = StatusFlags()
        undo = self.process_batch_command(command, f, batch)
        if undo.flags.success:
            self.process_flags(f)
            self.document.undo_stack_changed = True
        self.refresh_toolbar_state()
        return undo

    def process_batch_command(self, command, f, batch=None):
        """Process a single command but don't update the UI immediately.
        Instead, update the batch flags to reflect the changes needed to
        the UI.
        
        """
        undo = self.document.undo_stack.perform(command, self, batch)
        f.add_flags(undo.flags, command)
        return undo

    def process_flags(self, flags):
        """Perform the UI updates given the StatusFlags or BatchFlags flags
        
        """
        log.debug("processing flags: %s" % str(flags))
        d = self.document
        visible_range = False
        do_refresh = flags.refresh_needed

        if flags.cursor_index is not None:
            self.cursor_index = self.anchor_start_index = self.anchor_initial_start_index = self.anchor_end_index = self.anchor_initial_end_index = flags.cursor_index
            visible_range = True

        if flags.index_range is not None:
            if flags.select_range:
                self.anchor_start_index = self.anchor_initial_start_index = flags.index_range[0]
                self.anchor_end_index = self.anchor_initial_end_index = flags.index_range[1]
                d.change_count += 1
            visible_range = True

            # mark_index_range_changed should be called only when a value/style
            # changes, not if a range is simply selected.
            if flags.byte_values_changed or flags.byte_style_changed:
                self.mark_index_range_changed(flags.index_range)

        if flags.message:
            self.task.status_bar.message = flags.message

        if flags.metadata_dirty:
            self.metadata_dirty = True

        if flags.byte_values_changed:
            d.byte_values_changed = True  # also handles style changes and refresh
            do_refresh = False
        elif flags.byte_style_changed:
            d.byte_style_changed = True  # also handles refresh
            do_refresh = False

        if visible_range:
            # Only update the range on the current editor, not other views
            # which are allowed to remain where they are
            self.ensure_visible(self.anchor_start_index, self.anchor_end_index)

            # Prevent a double refresh since ensure_visible does a refresh as a
            # side effect.
            do_refresh = False

        if flags.rebuild_ui:
            self.rebuild_ui()
        elif do_refresh:
            self.refresh_panes()

    def popup_context_menu_from_commands(self, window, commands):
        """Popup a simple context menu with menu items defined by commands.
        
        Each entry is either None to indicate a separator, or a 2-tuple
        containing the string name and the command instance to process if the
        menu entry is selected.
        """
        popup = wx.Menu()
        context_menu_data = dict()
        for entry in commands:
            if entry is None:
                popup.AppendSeparator()
            else:
                name, cmd = entry
                i = wx.NewId()
                popup.Append(i, name)
                context_menu_data[i] = cmd
        ret = window.GetPopupMenuSelectionFromUser(popup)
        if ret == wx.ID_NONE:
            return
        cmd = context_menu_data[ret]
        self.process_command(cmd)

    def popup_context_menu_from_actions(self, window, actions, popup_data=None):
        """Popup a simple context menu with menu items defined by actions.
        
        Each entry is either None to indicate a separator, or an action to be
        used as the menu item.  Note that the action may be either an Action
        instance or the class of that Action.  If it is a class, a temporary
        instance of that class will be created.
        """
        popup = wx.Menu()
        context_menu_data = dict()

        def add_to_menu(menu, action, context_menu_data):
            i = wx.NewId()
            if not hasattr(action, 'task'):
                action = action(task=self.task)

            # wxpython popup entries can't have empty name
            name = action.name if action.name else " "
            item = menu.Append(i, name)
            item.Enable(action.enabled)
            context_menu_data[i] = action

        def loop_over_actions(popup, actions):
            for action in actions:
                if action is None:
                    popup.AppendSeparator()
                elif hasattr(action, '__iter__'):
                    submenu = wx.Menu()
                    title = action[0]
                    popup.AppendMenu(wx.NewId(), title, submenu)
                    loop_over_actions(submenu, action[1:])
                else:
                    add_to_menu(popup, action, context_menu_data)

        loop_over_actions(popup, actions)

        ret = window.GetPopupMenuSelectionFromUser(popup)
        if ret == wx.ID_NONE:
            return
        action = context_menu_data[ret]
        action_event = ActionEvent(task=self.task, popup_data=popup_data)
        action.perform(action_event)

    #### Traits event handlers

    # Trait event handlers are used instead of calling these functions directly
    # because multiple views of the document might exist and all editors will
    # be informed of the update via the trait event on the document.

    @on_trait_change('document:undo_stack_changed')
    def undo_stack_changed(self):
        log.debug("undo_stack_changed called!!!")
        self.update_undo_redo()
        self.update_history()
        wx.CallAfter(self.refresh_toolbar_state)

    @on_trait_change('document:byte_values_changed')
    def byte_values_changed(self):
        log.debug("byte_values_changed called!!!")
        self.document.change_count += 1
        self.invalidate_search()
        self.compare_to_baseline()
        self.rebuild_display_objects()
        self.refresh_panes()

    @on_trait_change('document:byte_style_changed')
    def byte_style_changed(self):
        log.debug("byte_style_changed called!!!")
        self.document.change_count += 1
        self.rebuild_display_objects()
        # styling can affect formatting, so rebuild display contents
        self.reconfigure_panes()

    #### wx hacks

    def refresh_toolbar_state(self):
        """The toolbar doesn't seem to refresh itself when its state is changed
        programmatically. This is provided as a convenience function to update
        any tools that are trait-dependent.
        """
        # Instead of calling AUI manager refresh (which was taking over 50% of
        # the refresh time) just refresh the toolbars themselves
        for pane in self.window._aui_manager._panes:
            if pane.IsToolbar():
                pane.window.Refresh()

    #### convenience functions

    @property
    def section_name(self):
        """If the document can be broken up into sections, this should return
        the title of the current section so it can be shown in the title bar of
        the window.
        """
        return ""

    @property
    def task(self):
        return self.editor_area.task

    @property
    def window(self):
        return self.editor_area.task.window

    @property
    def most_recent_uri(self):
        return self.document.uri



""" Examples for lazy adapter factories.

This module should be only imported when the adaptation takes place.
"""


class IBar(object):
    pass

class IBarToIFoo(object):
    pass

class IFoo(object):
    pass

#### EOF ######################################################################



"""
fs.tempfs
=========

Make a temporary file system that exists in a folder provided by the OS. All files contained in a TempFS are removed when the `close` method is called (or when the TempFS is cleaned up by Python).

"""

import os
import os.path
import time
import tempfile

from fs.base import synchronize
from fs.osfs import OSFS
from fs.errors import *

from fs import _thread_synchronize_default


class TempFS(OSFS):

    """Create a Filesystem in a temporary directory (with tempfile.mkdtemp),
    and removes it when the TempFS object is cleaned up."""

    _meta = dict(OSFS._meta)
    _meta['pickle_contents'] = False
    _meta['network'] = False
    _meta['atomic.move'] = True
    _meta['atomic.copy'] = True

    def __init__(self, identifier=None, temp_dir=None, dir_mode=0700, thread_synchronize=_thread_synchronize_default):
        """Creates a temporary Filesystem

        identifier -- A string that is included in the name of the temporary directory,
        default uses "TempFS"

        """
        self.identifier = identifier
        self.temp_dir = temp_dir
        self.dir_mode = dir_mode
        self._temp_dir = tempfile.mkdtemp(identifier or "TempFS", dir=temp_dir)
        self._cleaned = False
        super(TempFS, self).__init__(self._temp_dir, dir_mode=dir_mode, thread_synchronize=thread_synchronize)

    def __repr__(self):
        return '<TempFS: %s>' % self._temp_dir

    __str__ = __repr__

    def __unicode__(self):
        return u'<TempFS: %s>' % self._temp_dir

    def __getstate__(self):
        # If we are picking a TempFS, we want to preserve its contents,
        # so we *don't* do the clean
        state = super(TempFS, self).__getstate__()
        self._cleaned = True
        return state

    def __setstate__(self, state):
        state = super(TempFS, self).__setstate__(state)
        self._cleaned = False
        #self._temp_dir = tempfile.mkdtemp(self.identifier or "TempFS", dir=self.temp_dir)
        #super(TempFS, self).__init__(self._temp_dir,
        #                             dir_mode=self.dir_mode,
        #                             thread_synchronize=self.thread_synchronize)

    @synchronize
    def close(self):
        """Removes the temporary directory.

        This will be called automatically when the object is cleaned up by
        Python, although it is advisable to call it manually.
        Note that once this method has been called, the FS object may
        no longer be used.
        """
        super(TempFS, self).close()
        #  Depending on how resources are freed by the OS, there could
        #  be some transient errors when freeing a TempFS soon after it
        #  was used.  If they occur, do a small sleep and try again.
        try:
            self._close()
        except (ResourceLockedError, ResourceInvalidError):
            time.sleep(0.5)
            self._close()

    @convert_os_errors
    def _close(self):
        """Actual implementation of close().

        This is a separate method so it can be re-tried in the face of
        transient errors.
        """
        os_remove = convert_os_errors(os.remove)
        os_rmdir = convert_os_errors(os.rmdir)
        if not self._cleaned and self.exists("/"):
            self._lock.acquire()
            try:
                # shutil.rmtree doesn't handle long paths on win32,
                # so we walk the tree by hand.
                entries = os.walk(self.root_path, topdown=False)
                for (dir, dirnames, filenames) in entries:
                    for filename in filenames:
                        try:
                            os_remove(os.path.join(dir, filename))
                        except ResourceNotFoundError:
                            pass
                    for dirname in dirnames:
                        try:
                            os_rmdir(os.path.join(dir, dirname))
                        except ResourceNotFoundError:
                            pass
                try:
                    os.rmdir(self.root_path)
                except OSError:
                    pass
                self._cleaned = True
            finally:
                self._lock.release()
        super(TempFS, self).close()



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/13/2004
#
#------------------------------------------------------------------------------

""" Defines the ThemedCellRenderer class used to render theme-based cells for
    the TableEditor.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

import wx

from wx.grid \
    import PyGridCellRenderer, GridCellStringRenderer

from traitsui.ui_traits \
    import convert_bitmap

from helper \
    import BufferDC

#-------------------------------------------------------------------------
#  'ThemedCellRenderer' class:
#-------------------------------------------------------------------------


class ThemedCellRenderer(PyGridCellRenderer):
    """ Defines the ThemedCellRenderer class used to render theme-based cells
        for the TableEditor.
    """

    def __init__(self, column):
        """ Creates a new ThemedCellRenderer.
        """
        PyGridCellRenderer.__init__(self)

        # We are merging the pyface grid GridCellRenderer interface with the
        # wx.grid.PyGridCellRenderer interface, so we need to do this:
        self.renderer = self

        # Save the reference to the TableColumn:
        self.column = column

    #-- PyFace grid GridCellRenderer Method Overrides ------------------------

    # Invoked on left-button mouse clicks:
    def on_left_click(self, grid, row, col):
        return False

    # Invoked on left-button mouse double clicks:
    def on_left_dclick(self, grid, row, col):
        return False

    # Invoked on right-button mouse clicks:
    def on_right_click(self, grid, row, col):
        return False

    # Invoked on right-button mouse double clicks:
    def on_right_dclick(self, grid, row, col):
        return False

    # Invoked on a key press:
    def on_key(self, grid, row, col, key_event):
        return False

    # Clean-up:
    def dispose(self):
        del self.renderer
        del self.column

    #-- wx.GridCellRenderer Method Overrides ---------------------------------

    def Draw(self, grid, attr, dc, rect, row, col, is_selected):
        """ Draws the contents of the specified grid cell.
        """
        # Get the model object this cell is being rendered for:
        model = grid.grid.model
        object = model.get_filtered_item(row)

        # Get the draw bounds:
        x0 = rect.x
        y0 = rect.y
        dx = rect.width
        dy = rect.height

        # Do all drawing into an off-screen buffer:
        bdc = BufferDC(dc, dx, dy)

        # Draw the appropriate theme background:
        column = self.column
        if is_selected:
            theme = (column.get_selected_theme(object) or
                     column.get_cell_theme(object))
        else:
            theme = column.get_cell_theme(object)

        # If no column theme is specified, try to get the global theme from the
        # model:
        if theme is None:
            if row & 1:
                theme = model.alt_theme or model.cell_theme
            else:
                theme = model.cell_theme

            if is_selected:
                theme = model.selected_theme or theme

        if theme is not None:
            content = theme.content
            slice = theme.image_slice
            slice.fill(bdc, 0, 0, dx, dy)

            # Set up the correct text color to use:
            bdc.SetTextForeground(theme.content_color)

            # Calculate the margins for the draw area:
            left = slice.xleft + content.left
            top = slice.xtop + content.top
            right = slice.xright + content.right
            bottom = slice.xbottom + content.bottom
            ox, oy = theme.label.left, theme.label.top
        else:
            if is_selected:
                bg_color = grid.GetSelectionBackground()
            else:
                bg_color = attr.GetBackgroundColour()

            bdc.SetBackgroundMode(wx.SOLID)
            bdc.SetBrush(wx.Brush(bg_color, wx.SOLID))
            bdc.SetPen(wx.TRANSPARENT_PEN)
            bdc.DrawRectangle(0, 0, dx, dy)

            # Set up the correct text color to use:
            bdc.SetTextForeground(attr.GetTextColour())

            # Calculate the margins for the draw area:
            left = right = self.column.horizontal_margin
            top = bottom = self.column.vertical_margin
            ox = oy = 0

        # Get the alignment information:
        halign, valign = attr.GetAlignment()

        # Draw the bar graph (if any):
        maximum = column.get_maximum(object)
        if (not is_selected) and (maximum > 0.0):
            if theme is None:
                left = right = top = bottom = 0
            try:
                ratio = max(min(column.get_raw_value(object) / maximum,
                                1.0), -1.0)
                avail_dx = dx - left - right
                bar_dx = int(round(ratio * avail_dx))
                if halign == wx.ALIGN_CENTRE:
                    bar_dx /= 2
                    bar_x = left + (avail_dx / 2) + min(0, bar_dx)
                else:
                    bar_dx = abs(bar_dx)
                    if halign == wx.ALIGN_LEFT:
                        bar_x = left
                        left += self.column.horizontal_margin
                    else:
                        bar_x = avail_dx - bar_dx
                        right += self.column.horizontal_margin

                if bar_dx > 0:
                    bdc.SetBackgroundMode(wx.SOLID)
                    bdc.SetBrush(wx.Brush(column.get_graph_color(object),
                                          wx.SOLID))
                    bdc.SetPen(wx.TRANSPARENT_PEN)
                    bdc.DrawRectangle(bar_x, top, bar_dx, dy - top - bottom)
            except:
                pass

            if theme is None:
                left = right = self.column.horizontal_margin
                top = bottom = self.column.vertical_margin

        # Get the optional image bitmap and text:
        bitmap = convert_bitmap(column.get_image(object))
        text = grid.GetCellValue(row, col)

        # If no text or bitmap to display, then we are done:
        if (bitmap is None) and (text == ''):
            bdc.copy(x0, y0)
            return

        # Get the bitmap size:
        idx = idy = tdx = tdy = 0
        if bitmap is not None:
            idx = bitmap.GetWidth()
            idy = bitmap.GetHeight()

        # Get the text size:
        if text != '':
            bdc.SetFont(attr.GetFont())
            tdx, tdy = bdc.GetTextExtent(text)

            # Get the spacing between text and image:
            if bitmap is not None:
                idx += self.column.horizontal_margin

        # Calculate the x-coordinate of the image/text:
        if halign == wx.ALIGN_LEFT:
            x = left
        elif halign == wx.ALIGN_CENTRE:
            x = (left + ((dx - left - right - tdx - idx) / 2))
        else:
            x = (dx - right - tdx - idx)

        # Calculate the y-coordinate of the image/text:
        max_dy = max(tdy, idy)
        if valign == wx.ALIGN_TOP:
            y = top
        elif valign == wx.ALIGN_CENTRE:
            y = (top + ((dy - top - bottom - max_dy) / 2))
        else:
            y = (dy - bottom - max_dy)

        # Set up the clipping region to prevent drawing outside the margins:
        bdc.SetClippingRegion(left, top, dx - left - right, dy - top - bottom)

        # Draw the image (if left or center aligned):
        if (bitmap is not None) and (halign != wx.ALIGN_RIGHT):
            bdc.DrawBitmap(bitmap, x, y + ((max_dy - idy) / 2), True)
            x += idx

        # Finally, draw the text:
        if text != '':
            bdc.SetBackgroundMode(wx.TRANSPARENT)
            bdc.DrawText(text, x + ox, y + oy)
            x += tdx + self.column.horizontal_margin

        # Draw the image (if right-aligned):
        if (bitmap is not None) and (halign == wx.ALIGN_RIGHT):
            bdc.DrawBitmap(bitmap, x, y + ((max_dy - idy) / 2), True)

        # Discard the clipping region:
        bdc.DestroyClippingRegion()

        # Copy the buffer to the display:
        bdc.copy(x0, y0)

    def GetBestSize(self, grid, attr, dc, row, col):
        """ Determine best size for the cell. """
        # Get the model object this cell is being rendered for:
        object = grid.grid.model.get_filtered_item(row)

        # Get the text for this cell:
        text = grid.GetCellValue(row, col) or 'My'

        # Now calculate and return the best size for the text and image:
        dc.SetFont(attr.GetFont())
        tdx, tdy = dc.GetTextExtent(text)

        column = self.column
        bitmap = convert_bitmap(column.get_image(object))
        if bitmap is not None:
            tdx += (bitmap.GetWdth() + self.column.horizontal_margin)
            tdy = max(tdy, bitmap.GetHeight())

        theme = column.get_cell_theme(object)
        if theme is None:
            return wx.Size(
                tdx + self.column.horizontal_margin * 2,
                tdy + self.column.vertical_margin * 2)

        content = theme.content
        tdx += (content.left + content.right)
        tdy += (content.top + content.bottom)
        slice = theme.image_slice

        return wx.Size(max(slice.left + slice.right,
                           slice.xleft + slice.xright + tdx),
                       max(slice.top + slice.bottom,
                           slice.xtop + slice.xbottom + tdy))

    def Clone(self):
        return self.__class__(self.column)



""" Test data for testing the protocol manager with ABCs. """


from abc import ABCMeta

from traits.adaptation.api import PurePythonAdapter as Adapter


#### 'Power plugs' metaphor ###################################################

#### Protocols ################################################################

class UKStandard(object):
    __metaclass__ = ABCMeta

class EUStandard(object):
    __metaclass__ = ABCMeta

class JapanStandard(object):
    __metaclass__ = ABCMeta

class IraqStandard(object):
    __metaclass__ = ABCMeta

#### Implementations ##########################################################

class UKPlug(object):
    pass

UKStandard.register(UKPlug)

class EUPlug(object):
    pass

EUStandard.register(EUPlug)

class JapanPlug(object):
    pass

JapanStandard.register(JapanPlug)

class IraqPlug(object):
    pass

IraqStandard.register(IraqPlug)

class TravelPlug(object):

    def __init__(self, mode):
        self.mode = mode

#### Adapters #################################################################

# UK->EU
class UKStandardToEUStandard(Adapter):
    pass

EUStandard.register(UKStandardToEUStandard)

# EU->Japan
class EUStandardToJapanStandard(Adapter):
    pass

JapanStandard.register(EUStandardToJapanStandard)

# Japan->Iraq
class JapanStandardToIraqStandard(Adapter):
    pass

IraqStandard.register(JapanStandardToIraqStandard)

# EU->Iraq
class EUStandardToIraqStandard(Adapter):
    pass

IraqStandard.register(EUStandardToIraqStandard)

# UK->Japan
class UKStandardToJapanStandard(Adapter):
    pass

JapanStandard.register(UKStandardToJapanStandard)

# Travel->Japan
class TravelPlugToJapanStandard(Adapter):
    pass

JapanStandard.register(TravelPlugToJapanStandard)

# Travel->EU
class TravelPlugToEUStandard(Adapter):
    pass

EUStandard.register(TravelPlugToEUStandard)


#### 'Editor, Scriptable, Undoable' metaphor ##################################

class FileType(object):
    pass

class IEditor(object):
    __metaclass__ = ABCMeta

class IScriptable(object):
    __metaclass__ = ABCMeta

class IUndoable(object):
    __metaclass__ = ABCMeta


class FileTypeToIEditor(Adapter):
    pass

IEditor.register(FileTypeToIEditor)
IScriptable.register(FileTypeToIEditor)

class IScriptableToIUndoable(Adapter):
    pass

IUndoable.register(IScriptableToIUndoable)


#### Hierarchy example ########################################################

class IPrintable(object):
    __metaclass__ = ABCMeta

class Editor(object):
    pass

class TextEditor(Editor):
    pass

class EditorToIPrintable(Adapter):
    pass

IPrintable.register(EditorToIPrintable)

class TextEditorToIPrintable(Adapter):
    pass

IPrintable.register(TextEditorToIPrintable)


#### Interface hierarchy example ##############################################

class IPrimate(object):
    __metaclass__ = ABCMeta

class IHuman(IPrimate):
    pass

class IChild(IHuman):
    pass

class IIntermediate(object):
    __metaclass__ = ABCMeta

class ITarget(object):
    __metaclass__ = ABCMeta

class Source(object):
    pass

IChild.register(Source)

class IChildToIIntermediate(Adapter):
    pass

IIntermediate.register(IChildToIIntermediate)

class IHumanToIIntermediate(Adapter):
    pass

IIntermediate.register(IHumanToIIntermediate)

class IPrimateToIIntermediate(Adapter):
    pass

IIntermediate.register(IPrimateToIIntermediate)

class IIntermediateToITarget(Adapter):
    pass

ITarget.register(IIntermediateToITarget)


#### Non-trivial chaining example #############################################

class IStart(object):
    __metaclass__ = ABCMeta

class IGeneric(object):
    __metaclass__ = ABCMeta

class ISpecific(IGeneric):
    pass

class IEnd(object):
    __metaclass__ = ABCMeta

class Start(object):
    pass

IStart.register(Start)

class IStartToISpecific(Adapter):
    pass

ISpecific.register(IStartToISpecific)

class IGenericToIEnd(Adapter):
    pass

IEnd.register(IGenericToIEnd)

#### EOF ######################################################################



"""
fs.expose.ftp
==============

Expose an FS object over FTP (via pyftpdlib).

This module provides the necessary interfaces to expose an FS object over
FTP, plugging into the infrastructure provided by the 'pyftpdlib' module.

To use this in combination with fsserve, do the following:

$ fsserve -t 'ftp' $HOME

The above will serve your home directory in read-only mode via anonymous FTP on the
loopback address.
"""

import os
import stat
import time
import errno
from functools import wraps

from pyftpdlib import ftpserver

from fs.path import *
from fs.osfs import OSFS
from fs.errors import convert_fs_errors
from fs import iotools

from six import text_type as unicode


# Get these once so we can reuse them:
UID = os.getuid()
GID = os.getgid()


def decode_args(f):
    """
    Decodes string arguments using the decoding defined on the method's class.
    This decorator is for use on methods (functions which take a class or instance
    as the first parameter).

    Pyftpdlib (as of 0.7.0) uses str internally, so this decoding is necessary.
    """
    @wraps(f)
    def wrapper(self, *args):
        encoded = []
        for arg in args:
            if isinstance(arg, str):
                arg = arg.decode(self.encoding)
            encoded.append(arg)
        return f(self, *encoded)
    return wrapper


class FakeStat(object):
    """
    Pyftpdlib uses stat inside the library. This class emulates the standard
    os.stat_result class to make pyftpdlib happy. Think of it as a stat-like
    object ;-).
    """
    def __init__(self, **kwargs):
        for attr in dir(stat):
            if not attr.startswith('ST_'):
                continue
            attr = attr.lower()
            value = kwargs.get(attr, 0)
            setattr(self, attr, value)


class FTPFS(ftpserver.AbstractedFS):
    """
    The basic FTP Filesystem. This is a bridge between a pyfs filesystem and pyftpdlib's
    AbstractedFS. This class will cause the FTP server to serve the given fs instance.
    """
    encoding = 'utf8'
    "Sets the encoding to use for paths."

    def __init__(self, fs, root, cmd_channel, encoding=None):
        self.fs = fs
        if encoding is not None:
            self.encoding = encoding
        super(FTPFS, self).__init__(root, cmd_channel)

    def close(self):
        # Close and dereference the pyfs file system.
        if self.fs:
            self.fs.close()
        self.fs = None

    def validpath(self, path):
        try:
            normpath(path)
            return True
        except:
            return False

    @convert_fs_errors
    @decode_args
    @iotools.filelike_to_stream
    def open(self, path, mode, **kwargs):
        return self.fs.open(path, mode, **kwargs)

    @convert_fs_errors
    def chdir(self, path):
        # We dont' use the decorator here, we actually decode a version of the
        # path for use with pyfs, but keep the original for use with pyftpdlib.
        if not isinstance(path, unicode):
            # pyftpdlib 0.7.x
            unipath = unicode(path, self.encoding)
        else:
            # pyftpdlib 1.x
            unipath = path
        # TODO: can the following conditional checks be farmed out to the fs?
        # If we don't raise an error here for files, then the FTP server will
        # happily allow the client to CWD into a file. We really only want to
        # allow that for directories.
        if self.fs.isfile(unipath):
            raise OSError(errno.ENOTDIR, 'Not a directory')
        # similarly, if we don't check for existence, the FTP server will allow
        # the client to CWD into a non-existent directory.
        if not self.fs.exists(unipath):
            raise OSError(errno.ENOENT, 'Does not exist')
        # We use the original path here, so we don't corrupt self._cwd
        self._cwd = self.ftp2fs(path)

    @convert_fs_errors
    @decode_args
    def mkdir(self, path):
        self.fs.makedir(path)

    @convert_fs_errors
    @decode_args
    def listdir(self, path):
        return map(lambda x: x.encode(self.encoding), self.fs.listdir(path))

    @convert_fs_errors
    @decode_args
    def rmdir(self, path):
        self.fs.removedir(path)

    @convert_fs_errors
    @decode_args
    def remove(self, path):
        self.fs.remove(path)

    @convert_fs_errors
    @decode_args
    def rename(self, src, dst):
        self.fs.rename(src, dst)

    @convert_fs_errors
    @decode_args
    def chmod(self, path, mode):
        return

    @convert_fs_errors
    @decode_args
    def stat(self, path):
        info = self.fs.getinfo(path)
        kwargs = {
            'st_size': info.get('size'),
        }
        # Give the fs a chance to provide the uid/gid. Otherwise echo the current
        # uid/gid.
        kwargs['st_uid'] = info.get('st_uid', UID)
        kwargs['st_gid'] = info.get('st_gid', GID)
        if 'st_atime' in info:
            kwargs['st_atime'] = info['st_atime']
        elif 'accessed_time' in info:
            kwargs['st_atime'] = time.mktime(info["accessed_time"].timetuple())
        if 'st_mtime' in info:
            kwargs['st_mtime'] = info.get('st_mtime')
        elif 'modified_time' in info:
            kwargs['st_mtime'] = time.mktime(info["modified_time"].timetuple())
        # Pyftpdlib uses st_ctime on Windows platform, try to provide it.
        if 'st_ctime' in info:
            kwargs['st_ctime'] = info['st_ctime']
        elif 'created_time' in info:
            kwargs['st_ctime'] = time.mktime(info["created_time"].timetuple())
        elif 'st_mtime' in kwargs:
            # As a last resort, just copy the modified time.
            kwargs['st_ctime'] = kwargs['st_mtime']
        # Try to use existing mode.
        if 'st_mode' in info:
            kwargs['st_mode'] = info['st_mode']
        elif 'mode' in info:
            kwargs['st_mode'] = info['mode']
        else:
            # Otherwise, build one. Not executable by default.
            mode = 0660
            # Merge in the type (dir or file). File is tested first, some file systems
            # such as ArchiveMountFS treat archive files as directories too. By checking
            # file first, any such files will be only files (not directories).
            if self.fs.isfile(path):
                mode |= stat.S_IFREG
            elif self.fs.isdir(path):
                mode |= stat.S_IFDIR
                mode |= 0110  # Merge in exec bit to signal dir is listable
            kwargs['st_mode'] = mode
        return FakeStat(**kwargs)

    # No link support...
    lstat = stat

    @convert_fs_errors
    @decode_args
    def isfile(self, path):
        return self.fs.isfile(path)

    @convert_fs_errors
    @decode_args
    def isdir(self, path):
        return self.fs.isdir(path)

    @convert_fs_errors
    @decode_args
    def getsize(self, path):
        return self.fs.getsize(path)

    @convert_fs_errors
    @decode_args
    def getmtime(self, path):
        return self.stat(path).st_mtime

    def realpath(self, path):
        return path

    def lexists(self, path):
        return True


class FTPFSHandler(ftpserver.FTPHandler):
    """
    An FTPHandler class that closes the filesystem when done.
    """

    def close(self):
        # Close the FTPFS instance, it will close the pyfs file system.
        if self.fs:
            self.fs.close()
        super(FTPFSHandler, self).close()


class FTPFSFactory(object):
    """
    A factory class which can hold a reference to a file system object and
    encoding, then later pass it along to an FTPFS instance. An instance of
    this object allows multiple FTPFS instances to be created by pyftpdlib
    while sharing the same fs.
    """
    def __init__(self, fs, encoding=None):
        """
        Initializes the factory with an fs instance.
        """
        self.fs = fs
        self.encoding = encoding

    def __call__(self, root, cmd_channel):
        """
        This is the entry point of pyftpdlib. We will pass along the two parameters
        as well as the previously provided fs instance and encoding.
        """
        return FTPFS(self.fs, root, cmd_channel, encoding=self.encoding)


class HomeFTPFS(FTPFS):
    """
    A file system which serves a user's home directory.
    """
    def __init__(self, root, cmd_channel):
        """
        Use the provided user's home directory to create an FTPFS that serves an OSFS
        rooted at the home directory.
        """
        super(DemoFS, self).__init__(OSFS(root_path=root), '/', cmd_channel)


def serve_fs(fs, addr, port):
    """
    Creates a basic anonymous FTP server serving the given FS on the given address/port
    combo.
    """
    from pyftpdlib.contrib.authorizers import UnixAuthorizer
    ftp_handler = FTPFSHandler
    ftp_handler.authorizer = ftpserver.DummyAuthorizer()
    ftp_handler.authorizer.add_anonymous('/')
    ftp_handler.abstracted_fs = FTPFSFactory(fs)
    s = ftpserver.FTPServer((addr, port), ftp_handler)
    s.serve_forever()



# Standard library imports.
import unittest

# Enthought library imports.
from pyface.action.api import Action, ActionItem, ActionManager, Group, \
     MenuManager, MenuBarManager
from pyface.tasks.action.api import GroupSchema, MenuSchema, MenuBarSchema, \
     SchemaAddition
from pyface.tasks.action.task_action_manager_builder import \
     TaskActionManagerBuilder
from pyface.tasks.api import Task


class ActionManagerBuilderTestCase(unittest.TestCase):

    #### 'TestCase' protocol ##################################################

    def setUp(self):
        """ Create some dummy actions to use while testing.
        """
        for i in xrange(1, 7):
            action_id = 'action%i' % i
            setattr(self, action_id, Action(id=action_id, name='Action %i'%i))

    #### 'ActionManagerBuilderTestCase' protocol ##############################

    def assertActionElementsEqual(self, first, second):
        """ Checks that two action managers are (logically) equivalent.
        """
        children1 = children2 = []
        self.assertEquals(type(first), type(second))
        self.assertEquals(first.id, second.id)

        if isinstance(first, ActionItem):
            self.assertEquals(first.action.name, second.action.name)

        elif isinstance(first, ActionManager):
            if not isinstance(first, MenuBarManager):
                self.assertEquals(first.name, second.name)
            children1, children2 = first.groups, second.groups

        elif isinstance(first, Group):
            self.assertEquals(first.separator, second.separator)
            children1, children2 = first.items, second.items

        self.assertEquals(len(children1), len(children2))
        for i in xrange(len(children1)):
            self.assertActionElementsEqual(children1[i], children2[i])

    #### Tests ################################################################

    def test_simple_menu_bar(self):
        """ Does constructing a simple menu with no additions work?
        """
        schema = MenuBarSchema(
            MenuSchema(self.action1, self.action2, id='File', name='&File'),
            MenuSchema(self.action3, self.action4, id='Edit', name='&Edit'))
        builder = TaskActionManagerBuilder(task=Task(menu_bar=schema))
        actual = builder.create_menu_bar_manager()
        desired = MenuBarManager(MenuManager(self.action1, self.action2,
                                             id='File', name='&File'),
                                 MenuManager(self.action3, self.action4,
                                             id='Edit', name='&Edit'),
                                 id='MenuBar')
        self.assertActionElementsEqual(actual, desired)

    #### Tests about schema additions #########################################

    def test_additions_menu_bar(self):
        """ Does constructing a menu with a few additions work?
        """
        schema = MenuBarSchema(
            MenuSchema(GroupSchema(self.action1, self.action2, id='FileGroup'),
                       id='File'))
        extras = [ SchemaAddition(factory=lambda: self.action3,
                                  before='action1',
                                  path='MenuBar/File/FileGroup'),
                   SchemaAddition(factory=lambda: self.action4,
                                  before='action1',
                                  path='MenuBar/File/FileGroup'),
                   SchemaAddition(factory=lambda: self.action5,
                                  path='MenuBar/File/FileGroup')]
        builder = TaskActionManagerBuilder(task=Task(menu_bar=schema,
                                                     extra_actions=extras))
        actual = builder.create_menu_bar_manager()
        desired = MenuBarManager(MenuManager(Group(self.action3, self.action4,
                                                   self.action1, self.action2,
                                                   self.action5,
                                                   id='FileGroup'),
                                             id='File'),
                                 id='MenuBar')
        self.assertActionElementsEqual(actual, desired)

    def test_extra_menu(self):
        """ Test contributing a whole new menu to the menu bar. """

        # Initial menu.
        schema = MenuBarSchema(
            MenuSchema(GroupSchema(self.action1, id='FileGroup'),
                       id='FileMenu')
        )

        # Contributed menu.
        extra_menu = MenuSchema(
            GroupSchema(self.action2, id='BarGroup'),
            id= 'DummyActionsMenu',
        )

        extra_actions = [
            SchemaAddition(path='MenuBar',
                           factory=lambda : extra_menu,
                           id='DummyActionsSMenu'),
        ]

        # Build the final menu.
        builder = TaskActionManagerBuilder(
            task=Task(menu_bar=schema, extra_actions=extra_actions)
        )
        actual = builder.create_menu_bar_manager()

        desired = MenuBarManager(
            MenuManager(Group(self.action1, id='FileGroup'),
                        id='FileMenu'),
            MenuManager(Group(self.action2, id='BarGroup'),
                        id='DummyActionsMenu'),
            id='MenuBar'
        )

        self.assertActionElementsEqual(actual, desired)

    #### Tests about merging schemas ##########################################

    def test_merging_redundant_items(self):
        """ Menus and groups with matching path are merged together. """

        # Initial menu.
        schema = MenuBarSchema(
            MenuSchema(GroupSchema(self.action1, id='FileGroup'),
                       name='File menu number one', id='FileMenu')
        )

        # Contributed menus.
        extra_menu = MenuSchema(
            GroupSchema(self.action2, id='FileGroup'),
            name='File menu number two',
            id= 'FileMenu',
        )

        extra_actions = [
            SchemaAddition(path='MenuBar',
                           factory=lambda : extra_menu,
                           id='DummyActionsSMenu'),
        ]

        # Build the final menu.
        builder = TaskActionManagerBuilder(
            task=Task(menu_bar=schema, extra_actions=extra_actions)
        )
        actual = builder.create_menu_bar_manager()

        # Note that we expect the name of the menu to be inherited from
        # the menu in the menu bar schema that is defined first.
        desired = MenuBarManager(
            MenuManager(Group(self.action1, self.action2, id='FileGroup'),
                        name='File menu number one', id='FileMenu'),
            id='MenuBar'
        )
        self.assertActionElementsEqual(actual, desired)

    def test_unwanted_merge(self):
        """ Test that we don't have automatic merges due to forgetting to set
        a schema ID. """

        # Initial menu.
        schema = MenuBarSchema(
            MenuSchema(GroupSchema(self.action1, id='FileGroup'),
                       name='File 1')
        )

        # Contributed menus.
        extra_menu = MenuSchema(
            GroupSchema(self.action2, id='FileGroup'),
            name='File 2'
        )

        extra_actions = [
            SchemaAddition(path='MenuBar',
                           factory=lambda : extra_menu,
                           id='DummyActionsSMenu'),
            ]

        # Build the final menu.
        builder = TaskActionManagerBuilder(
            task=Task(menu_bar=schema, extra_actions=extra_actions)
        )
        actual = builder.create_menu_bar_manager()

        # Note that we expect the name of the menu to be inherited from
        # the menu in the menu bar schema that is defined first.
        desired = MenuBarManager(
            MenuManager(Group(self.action1, id='FileGroup'),
                        name='File 1', id='MenuSchema_1'),
            MenuManager(Group(self.action2, id='FileGroup'),
                        name='File 2', id='MenuSchema_2'),
            id='MenuBar'
        )
        self.assertActionElementsEqual(actual, desired)

    def test_merging_items_with_same_id_but_different_class(self):
        """ Schemas with the same path but different types (menus, groups)
        are not merged together.

        Having a group and a menu with the same path is of course bad practice,
        but we need a predictable outcome.

        """

        # Initial menu.
        schema = MenuBarSchema(
            MenuSchema(GroupSchema(self.action1, id='FileGroup'),
                       id='FileSchema')
        )

        # Contributed menus.
        extra_group = GroupSchema(self.action2, id='FileSchema')

        extra_actions = [
            SchemaAddition(path='MenuBar',
                           factory=(lambda : extra_group),
                           id='DummyActionsSMenu'),
        ]

        # Build the final menu.
        builder = TaskActionManagerBuilder(
            task=Task(menu_bar=schema, extra_actions=extra_actions)
        )
        actual = builder.create_menu_bar_manager()

        desired = MenuBarManager(
            MenuManager(Group(self.action1, id='FileGroup'),
                        id='FileSchema'),
            Group(self.action2, id='FileSchema'),
            id='MenuBar'
        )
        self.assertActionElementsEqual(actual, desired)

    def test_merging_redundant_items_that_are_not_schemas(self):
        """ Items that are not schemas cannot be merged, but we should
        not crash, either. """

        # Initial menu.
        schema = MenuBarSchema(
            # This menu is not a schema...
            MenuManager(Group(self.action1, id='FileGroup'),
                        id='FileMenu')
        )

        # Contributed menus.
        extra_menu = MenuSchema(
            GroupSchema(self.action2, id='FileGroup'),
            id= 'FileMenu',
        )

        extra_actions = [
            SchemaAddition(path='MenuBar',
                           factory=lambda : extra_menu,
                           id='DummyActionsSMenu'),
        ]

        # Build the final menu.
        builder = TaskActionManagerBuilder(
            task=Task(menu_bar=schema, extra_actions=extra_actions)
        )
        actual = builder.create_menu_bar_manager()

        desired = MenuBarManager(
            MenuManager(Group(self.action1, id='FileGroup'),
                        id='FileMenu'),
            MenuManager(Group(self.action2, id='FileGroup'),
                        id='FileMenu'),
            id='MenuBar'
        )
        self.assertActionElementsEqual(actual, desired)

    #### Tests about ordering #################################################

    def test_absolute_ordering(self):
        """ Does specifying absolute_position work?
        """
        schema = MenuBarSchema(
            MenuSchema(GroupSchema(self.action1, self.action2, id='FileGroup'),
                       id='File'))
        extras = [ SchemaAddition(factory=lambda: self.action3,
                                  absolute_position='last',
                                  path='MenuBar/File/FileGroup'),
                   SchemaAddition(factory=lambda: self.action4,
                                  absolute_position='first',
                                  path='MenuBar/File/FileGroup'),
                   SchemaAddition(factory=lambda: self.action5,
                                  absolute_position='first',
                                  path='MenuBar/File/FileGroup')]
        builder = TaskActionManagerBuilder(task=Task(menu_bar=schema,
                                                     extra_actions=extras))
        actual = builder.create_menu_bar_manager()
        desired = MenuBarManager(MenuManager(Group(self.action4, self.action5,
                                                   self.action1, self.action2,
                                                   self.action3,
                                                   id='FileGroup'),
                                             id='File'),
                                 id='MenuBar')
        self.assertActionElementsEqual(actual, desired)

    def test_absolute_and_before_after(self):
        """ Does specifying absolute_position along with before, after work?
        """
        schema = MenuBarSchema(
            MenuSchema(GroupSchema(self.action1, self.action2, id='FileGroup'),
                       id='File'))
        extras = [ SchemaAddition(factory=lambda: self.action3,
                                  id='action3',
                                  after='action2',
                                  path='MenuBar/File/FileGroup'),
                   SchemaAddition(factory=lambda: self.action4,
                                  after='action3',
                                  path='MenuBar/File/FileGroup'),
                   SchemaAddition(factory=lambda: self.action5,
                                  id='action5',
                                  absolute_position='last',
                                  path='MenuBar/File/FileGroup'),
                   SchemaAddition(factory=lambda: self.action6,
                                  absolute_position='last',
                                  before='action5',
                                  path='MenuBar/File/FileGroup')
                   ]
        builder = TaskActionManagerBuilder(task=Task(menu_bar=schema,
                                                     extra_actions=extras))
        actual = builder.create_menu_bar_manager()
        desired = MenuBarManager(MenuManager(Group(self.action1, self.action2,
                                                   self.action3, self.action4,
                                                   self.action6, self.action5,
                                                   id='FileGroup'),
                                             id='File'),
                                 id='MenuBar')
        self.assertActionElementsEqual(actual, desired)

if __name__ == '__main__':
    unittest.main()



""" The service registry interface. """


# Enthought library imports.
from traits.api import Event, Interface


class IServiceRegistry(Interface):
    """ The service registry interface.

    The service registry provides a 'Yellow Pages' style mechanism, in that
    services are published and looked up by protocol (meaning, *interface*,
    *type*, or *class* (for old-style classes!). It is called a 'Yellow Pages'
    mechanism because it is just like looking up a telephone number in the
    'Yellow Pages' phone book. You use the 'Yellow Pages' instead of the
    'White Pages' when you don't know the *name* of the person you want to
    call but you do know what *kind* of service you require. For example, if
    you have a leaking pipe, you know you need a plumber, so you pick up your
    'Yellow Pages', go to the 'Plumbers' section and choose one that seems to
    fit the bill based on price, location, certification, etc. The service
    registry does exactly the same thing as the 'Yellow Pages', only with
    objects, and it even allows you to publish your own entries for free
    (unlike the *real* one)!

    """

    # An event that is fired when a service is registered.
    registered = Event

    # An event that is fired when a service is unregistered.
    unregistered = Event

    def get_service(self, protocol, query='', minimize='', maximize=''):
        """ Return at most one service that matches the specified query.

        The protocol can be an actual class or interface, or the *name* of a
        class or interface in the form '<module_name>.<class_name>'.

        Return None if no such service is found.

        If no query is specified then a service that provides the specified
        protocol is returned (if one exists).

        NOTE: If more than one service exists that match the criteria then
        Don't try to guess *which* one it will return - it is random!

        """

    def get_service_from_id(self, service_id):
        """ Return the service with the specified id.

        If no such service exists a 'ValueError' exception is raised.

        """

    def get_services(self, protocol, query='', minimize='', maximize=''):
        """ Return all services that match the specified query.

        The protocol can be an actual class or interface, or the *name* of a
        class or interface in the form '<module_name>.<class_name>'.

        If no services match the query, then an empty list is returned.

        If no query is specified then all services that provide the specified
        protocol are returned (if any exist).

        """

    def get_service_properties(self, service_id):
        """ Return the dictionary of properties associated with a service.

        If no such service exists a 'ValueError' exception is raised.

        The properties returned are 'live' i.e. changing them immediately
        changes the service registration.

        """

    def register_service(self, protocol, obj, properties=None):
        """ Register a service.

        The protocol can be an actual class or interface, or the *name* of a
        class or interface in the form::

            'foo.bar.baz'

        Which is turned into the equivalent of an import statement that
        looks like::

            from foo.bar import baz

        Return a service Id that can be used to unregister the service and to
        get/set any service properties.

        If 'obj' does not implement the specified protocol then it is treated
        as a 'service factory' that will be called the first time a service of
        the appropriate type is requested. A 'service factory' is simply a
        callable that takes the properties specified here as keyword arguments
        and returns an object. For *really* lazy loading, the factory can also
        be specified as a string which is used to import the callable.

        """

    def set_service_properties(self, service_id, properties):
        """ Set the dictionary of properties associated with a service.

        If no such service exists a 'ValueError' exception is raised.

        """

    def unregister_service(self, service_id):
        """ Unregister a service.

        If no such service exists a 'ValueError' exception is raised.

        """

#### EOF ######################################################################



from __future__ import absolute_import

from traits.testing.unittest_tools import UnittestTools, unittest

from ..action import Action
from ..action_item import ActionItem
from ..group import Group


class TestActionItem(unittest.TestCase, UnittestTools):

    def setUp(self):
        # test whether function is called by updating list
        # XXX should really use mock
        self.memo = []

        def perform():
            self.memo.append('called')

        self.perform = perform

        self.action = Action(name='Test', on_perform=perform)
        self.action_item = ActionItem(action=self.action)

    def test_init_action_item(self):
        group = Group(self.action_item)
        self.assertEqual(group.items, [self.action_item])
        self.assertEqual(self.action_item.parent, group)

    def test_init_action(self):
        group = Group(self.action)
        self.assertEqual(len(group.items), 1)
        self.assertEqual(group.items[0].action, self.action)
        self.assertEqual(group.items[0].parent, group)

    def test_init_callable(self):
        group = Group(self.perform)
        self.assertEqual(len(group.items), 1)
        self.assertEqual(group.items[0].action.on_perform, self.perform)
        self.assertEqual(group.items[0].action.name, "Perform")
        self.assertEqual(group.items[0].parent, group)

    def test_init_nothing(self):
        group = Group()
        self.assertEqual(group.items, [])

    def test_append(self):
        group = Group(self.action_item)
        action_item2 = ActionItem(action=Action(name='Action 2'))
        # XXX items doesn't fire a change event.  Should it?
        group.append(action_item2)
        self.assertEqual(group.items, [self.action_item, action_item2])
        self.assertEqual(action_item2.parent, group)

    def test_append_action(self):
        group = Group(self.action_item)
        action2 = Action(name='Action 2')
        # XXX items doesn't fire a change event.  Should it?
        group.append(action2)
        self.assertEqual(len(group.items), 2)
        self.assertEqual(group.items[0], self.action_item)
        self.assertEqual(group.items[1].action, action2)
        self.assertEqual(group.items[1].parent, group)

    def test_append_callable(self):
        group = Group(self.action_item)
        action2 = Action(name='Action 2')
        # XXX items doesn't fire a change event.  Should it?
        group.append(self.perform)
        self.assertEqual(len(group.items), 2)
        self.assertEqual(group.items[0], self.action_item)
        self.assertEqual(group.items[1].action.on_perform, self.perform)
        self.assertEqual(group.items[1].action.name, "Perform")
        self.assertEqual(group.items[1].parent, group)

    def test_clear(self):
        group = Group(self.action_item)
        # XXX items doesn't fire a change event.  Should it?
        group.clear()
        self.assertEqual(group.items, [])
        # XXX clear doesn't set items' parent to None, but remove does...
        # self.assertIsNone(self.action_item.parent)

    def test_destroy(self):
        group = Group(self.action_item)
        # XXX items doesn't fire a change event.  Should it?
        # XXX should mock action item's to ensure that destroy is called
        group.destroy()
        self.assertEqual(group.items, [self.action_item])

    def test_insert(self):
        group = Group(self.action_item)
        action_item2 = ActionItem(action=Action(name='Action 2'))
        # XXX items doesn't fire a change event.  Should it?
        group.insert(1, action_item2)
        self.assertEqual(group.items, [self.action_item, action_item2])
        self.assertEqual(action_item2.parent, group)

    def test_insert_action(self):
        group = Group(self.action_item)
        action2 = Action(name='Action 2')
        # XXX items doesn't fire a change event.  Should it?
        group.insert(1, action2)
        self.assertEqual(len(group.items), 2)
        self.assertEqual(group.items[0], self.action_item)
        self.assertEqual(group.items[1].action, action2)
        self.assertEqual(group.items[1].parent, group)

    def test_insert_callable(self):
        group = Group(self.action_item)
        action2 = Action(name='Action 2')
        # XXX items doesn't fire a change event.  Should it?
        group.insert(1, self.perform)
        self.assertEqual(len(group.items), 2)
        self.assertEqual(group.items[0], self.action_item)
        self.assertEqual(group.items[1].action.on_perform, self.perform)
        self.assertEqual(group.items[1].action.name, "Perform")
        self.assertEqual(group.items[1].parent, group)

    def test_insert_at_start(self):
        group = Group(self.action_item)
        action_item2 = ActionItem(action=Action(name='Action 2'))
        # XXX items doesn't fire a change event.  Should it?
        group.insert(0, action_item2)
        self.assertEqual(group.items, [action_item2, self.action_item])
        self.assertEqual(action_item2.parent, group)

    def test_remove(self):
        group = Group(self.action_item)
        # XXX items doesn't fire a change event.  Should it?
        group.remove(self.action_item)
        self.assertEqual(group.items, [])
        self.assertIsNone(self.action_item.parent)

    def test_remove_missing(self):
        group = Group()
        with self.assertRaises(ValueError):
            group.remove(self.action_item)

    def test_insert_before(self):
        group = Group(self.action_item)
        action_item2 = ActionItem(action=Action(name='Action 2'))
        # XXX items doesn't fire a change event.  Should it?
        group.insert_before(self.action_item, action_item2)
        self.assertEqual(group.items, [action_item2, self.action_item])
        self.assertEqual(action_item2.parent, group)

    def test_insert_after(self):
        group = Group(self.action_item)
        action_item2 = ActionItem(action=Action(name='Action 2'))
        # XXX items doesn't fire a change event.  Should it?
        group.insert_after(self.action_item, action_item2)
        self.assertEqual(group.items, [self.action_item, action_item2])
        self.assertEqual(action_item2.parent, group)

    def test_find(self):
        group = Group(self.action_item)
        item = group.find('Test')
        self.assertEqual(item, self.action_item)

    def test_find_missing(self):
        group = Group(self.action_item)
        item = group.find('Not here')
        self.assertIsNone(item)

    def test_enabled_changed(self):
        group = Group(self.action_item)
        group.enabled = False
        self.assertFalse(self.action_item.enabled)
        self.assertFalse(self.action.enabled)
        group.enabled = True
        self.assertTrue(self.action_item.enabled)
        self.assertTrue(self.action.enabled)



#------------------------------------------------------------------------------
#
#  Copyright (c) 2008, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/21/2004
#
#------------------------------------------------------------------------------
""" Defines the list editor factory for the traits user interface toolkits..
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import (
    HasTraits,
    BaseTraitHandler,
    Range,
    Str,
    Any,
    Int,
    Instance,
    Property,
    Bool,
    Callable,
    Enum,
    PrototypedFrom)

# CIRCULAR IMPORT FIXME: Importing from the source rather than traits.ui.api
# to avoid circular imports, as this EditorFactory will be part of
# traits.ui.api as well.
from ..view import View

from ..item import Item

from ..ui_traits import style_trait, AView

from ..editor_factory import EditorFactory

from ..toolkit import toolkit_object

# Currently, this traits is used only for the wx backend.
from ..helper import DockStyle

#-------------------------------------------------------------------------
#  Trait definitions:
#-------------------------------------------------------------------------

# Trait whose value is a BaseTraitHandler object
handler_trait = Instance(BaseTraitHandler)

# The visible number of rows displayed
rows_trait = Range(1, 50, 5,
                   desc='the number of list rows to display')

# The visible number of columns displayed
columns_trait = Range(1, 10, 1,
                      desc='the number of list columns to display')

editor_trait = Instance(EditorFactory)

#-------------------------------------------------------------------------
#  'ToolkitEditorFactory' class:
#-------------------------------------------------------------------------


class ToolkitEditorFactory(EditorFactory):
    """ Editor factory for list editors.
    """

    #-------------------------------------------------------------------------
    #  Trait definitions:
    #-------------------------------------------------------------------------

    # The editor to use for each list item:
    editor = editor_trait

    # Can the list be reorganized, or have items added and deleted.
    mutable = Bool(True)

    # The style of editor to use for each item:
    style = style_trait

    # The trait handler for each list item:
    trait_handler = handler_trait

    # The number of list rows to display:
    rows = rows_trait

    # The number of list columns to create:
    columns = columns_trait

    # Use a notebook for a custom view?
    use_notebook = Bool(False)

    # Show a right-click context menu for the notebook tabs?  (Qt only)
    show_notebook_menu = Bool(False)

    #-- Notebook Specific Traits ---------------------------------------------

    # Are notebook items deletable?
    deletable = Bool(False)

    # The extended name of the trait on each page object which should be used
    # to determine whether or not an individual page should be deletable.
    deletable_trait = Str()

    # FIXME: Currently, this trait is used only in the wx backend.
    # The DockWindow graphical theme
    dock_theme = Any

    # FIXME: Currently, this trait is used only in the wx backend.
    # Dock page style to use for each DockControl:
    dock_style = DockStyle

    # Export class for each item in a notebook:
    export = Str

    # Name of the view to use in notebook mode:
    view = AView

    # The type of UI to construct ('panel', 'subpanel', etc)
    ui_kind = Enum('subpanel', 'panel')

    # A factory function that can be used to define that actual object to be
    # edited (i.e. view_object = factory( object )):
    factory = Callable

    # Extended name to use for each notebook page. It can be either the actual
    # name or the name of an attribute on the object in the form:
    # '.name[.name...]'
    page_name = Str

    # Name of the [object.]trait[.trait...] to synchronize notebook page
    # selection with:
    selected = Str

    #-------------------------------------------------------------------------
    #  Traits view definition:
    #-------------------------------------------------------------------------

    traits_view = View([['use_notebook{Use a notebook in a custom view}',
                         '|[Style]'],
                        [Item('page_name',
                              enabled_when='object.use_notebook'),
                         Item('view',
                              enabled_when='object.use_notebook'),
                         '|[Notebook options]'],
                        [Item('rows',
                              enabled_when='not object.use_notebook'),
                         '|[Number of list rows to display]<>']])

    #-------------------------------------------------------------------------
    #  'Editor' factory methods:
    #-------------------------------------------------------------------------

    def _get_custom_editor_class(self):
        if self.use_notebook:
            return toolkit_object('list_editor:NotebookEditor')
        return toolkit_object('list_editor:CustomEditor')

#-------------------------------------------------------------------------
#  'ListItemProxy' class:
#   This class is used to update the list editors when the object changes
#   external to the editor.
#-------------------------------------------------------------------------


class ListItemProxy(HasTraits):

    # The list proxy:
    list = Property

    # The item proxies index into the original list:
    index = Int

    # Delegate all other traits to the original object:
    _ = PrototypedFrom('_zzz_object')

    # Define all of the private internal use values (the funny names are an
    # attempt to avoid name collisions with delegated trait names):
    _zzz_inited = Any
    _zzz_object = Any
    _zzz_name = Any

    def __init__(self, object, name, index, trait, value):
        super(ListItemProxy, self).__init__()

        self._zzz_inited = False
        self._zzz_object = object
        self._zzz_name = name
        self.index = index

        if trait is not None:
            self.add_trait('value', trait)
            self.value = value

        self._zzz_inited = (self.index < len(self.list))

    def _get_list(self):
        return getattr(self._zzz_object, self._zzz_name)

    def _value_changed(self, old_value, new_value):
        if self._zzz_inited:
            self.list[self.index] = new_value


# Define the ListEditor class
ListEditor = ToolkitEditorFactory

### EOF ---------------------------------------------------------------------



#!/usr/bin/env python
from fs.utils import copyfile, copyfile_non_atomic
from fs.path import pathjoin, iswildcard
from fs.commands.runner import Command
import sys
import Queue as queue
import time
import threading


class FileOpThread(threading.Thread):            
    
    def __init__(self, action, name, dest_fs, queue, on_done, on_error):
        self.action = action                  
        self.dest_fs = dest_fs
        self.queue = queue
        self.on_done = on_done
        self.on_error = on_error
        self.finish_event = threading.Event()
        super(FileOpThread, self).__init__()        
    
    def run(self):        
        
        while not self.finish_event.isSet():            
            try:
                path_type, fs, path, dest_path = self.queue.get(timeout=0.1)
            except queue.Empty:
                continue
            try:                
                if path_type == FScp.DIR:                        
                    self.dest_fs.makedir(path, recursive=True, allow_recreate=True)
                else:                                                                
                    self.action(fs, path, self.dest_fs, dest_path, overwrite=True)                    
            except Exception, e:                
                self.on_error(e)                                
                self.queue.task_done()                                  
                break                   
            else:
                self.queue.task_done() 
                self.on_done(path_type, fs, path, self.dest_fs, dest_path)
                      
                     

class FScp(Command):
    
    DIR, FILE = 0, 1
    
    usage = """fscp [OPTION]... [SOURCE]... [DESTINATION]
Copy SOURCE to DESTINATION"""
    
    def get_action(self):
        if self.options.threads > 1:
            return copyfile_non_atomic
        else:
            return copyfile
    
    def get_verb(self):
        return 'copying...'

    def get_optparse(self):
        optparse = super(FScp, self).get_optparse()
        optparse.add_option('-p', '--progress', dest='progress', action="store_true", default=False,
                            help="show progress", metavar="PROGRESS")
        optparse.add_option('-t', '--threads', dest='threads', action="store", default=1,
                            help="number of threads to use", type="int", metavar="THREAD_COUNT")        
        return optparse
        
    def do_run(self, options, args):
               
        self.options = options 
        if len(args) < 2:
            self.error("at least two filesystems required\n")
            return 1
        
        srcs = args[:-1]
        dst = args[-1] 
        
        dst_fs, dst_path = self.open_fs(dst, writeable=True, create_dir=True)
        
        if dst_path is not None and dst_fs.isfile(dst_path):
            self.error('Destination must be a directory\n')
            return 1
        
        if dst_path:
            dst_fs = dst_fs.makeopendir(dst_path)
            dst_path = None                      
                
        copy_fs_paths = []
        
        progress = options.progress                
        
        if progress:
            sys.stdout.write(self.progress_bar(len(srcs), 0, 'scanning...'))
            sys.stdout.flush()
        
        self.root_dirs = [] 
        for i, fs_url in enumerate(srcs):
            src_fs, src_path = self.open_fs(fs_url)                      

            if src_path is None:
                src_path = '/'

            if iswildcard(src_path):
                for file_path in src_fs.listdir(wildcard=src_path, full=True):
                    copy_fs_paths.append((self.FILE, src_fs, file_path, file_path))
                    
            else:                
                if src_fs.isdir(src_path): 
                    self.root_dirs.append((src_fs, src_path))                                        
                    src_sub_fs = src_fs.opendir(src_path)
                    for dir_path, file_paths in src_sub_fs.walk():
                        if dir_path not in ('', '/'):                                            
                            copy_fs_paths.append((self.DIR, src_sub_fs, dir_path, dir_path))
                        sub_fs = src_sub_fs.opendir(dir_path)
                        for file_path in file_paths:                                                         
                            copy_fs_paths.append((self.FILE, sub_fs, file_path, pathjoin(dir_path, file_path)))
                else:
                    if src_fs.exists(src_path):
                        copy_fs_paths.append((self.FILE, src_fs, src_path, src_path))
                    else:
                        self.error('%s is not a file or directory\n' % src_path)
                        return 1 
                    
            if progress:
                sys.stdout.write(self.progress_bar(len(srcs), i + 1, 'scanning...'))
                sys.stdout.flush()
                                       
        if progress:
            sys.stdout.write(self.progress_bar(len(copy_fs_paths), 0, self.get_verb()))
            sys.stdout.flush()
                                        
        if self.options.threads > 1:            
            copy_fs_dirs = [r for r in copy_fs_paths if r[0] == self.DIR]
            copy_fs_paths = [r for r in copy_fs_paths if r[0] == self.FILE]            
            for path_type, fs, path, dest_path in copy_fs_dirs:               
                dst_fs.makedir(path, allow_recreate=True, recursive=True)                             
        
        self.lock = threading.RLock()
            
        self.total_files = len(copy_fs_paths)
        self.done_files = 0
                
        file_queue = queue.Queue()        
        threads = [FileOpThread(self.get_action(),
                                'T%i' % i,
                                dst_fs,
                                file_queue,
                                self.on_done,
                                self.on_error)
                        for i in xrange(options.threads)]
        
        for thread in threads:
            thread.start()
        
        self.action_errors = []
        complete = False
        try:        
            enqueue = file_queue.put            
            for resource in copy_fs_paths:
                enqueue(resource)
                        
            while not file_queue.empty():
                time.sleep(0)
                if self.any_error():
                    raise SystemExit
            # Can't use queue.join here, or KeyboardInterrupt will not be
            # caught until the queue is finished 
            #file_queue.join()
        
        except KeyboardInterrupt:            
            options.progress = False                    
            self.output("\nCancelling...\n")
                
        except SystemExit:
            options.progress = False            
                               
        finally:
            sys.stdout.flush()                
            for thread in threads:
                thread.finish_event.set()                           
            for thread in threads:
                thread.join()
            complete = True
            if not self.any_error():
                self.post_actions()
                                   
        dst_fs.close()
        
        if self.action_errors:
            for error in self.action_errors:
                self.error(self.wrap_error(unicode(error)) + '\n')            
            sys.stdout.flush()
        else:
            if complete and options.progress:
                sys.stdout.write(self.progress_bar(self.total_files, self.done_files, ''))
                sys.stdout.write('\n')
                sys.stdout.flush()
        
    def post_actions(self):
        pass
        
    def on_done(self, path_type, src_fs, src_path, dst_fs, dst_path):        
        self.lock.acquire()        
        try:
            if self.options.verbose:
                if path_type == self.DIR:
                    print "mkdir %s" % dst_fs.desc(dst_path)
                else:
                    print "%s -> %s" % (src_fs.desc(src_path), dst_fs.desc(dst_path))
            elif self.options.progress:
                self.done_files += 1        
                sys.stdout.write(self.progress_bar(self.total_files, self.done_files, self.get_verb()))
                sys.stdout.flush()
        finally:
            self.lock.release()
            
    def on_error(self, e):        
        self.lock.acquire()
        try:
            self.action_errors.append(e)
        finally:
            self.lock.release()
    
    def any_error(self):                
        self.lock.acquire()
        try:
            return bool(self.action_errors)
        finally:
            self.lock.release()
            
    def progress_bar(self, total, remaining, msg=''):
        bar_width = 20
        throbber = '|/-\\'
        throb = throbber[remaining % len(throbber)]
        done = float(remaining) / total
        
        done_steps = int(done * bar_width)
        bar_steps = ('#' * done_steps).ljust(bar_width) 
                
        msg = '%s %i%%' % (msg, int(done * 100.0))        
        msg = msg.ljust(20)
        
        if total == remaining:
            throb = ''
            
        bar = '\r%s[%s] %s\r' % (throb, bar_steps, msg.lstrip())
        return bar
        
def run():
    return FScp().run()

if __name__ == "__main__":
    sys.exit(run())
